import torch
import torch.nn.functional as F


def confusion_mat(
    yhat: torch.Tensor, target: torch.Tensor, activation: str = None
) -> torch.Tensor:
    """Compute the confusion matrix from the soft mask and target tensor.

    Parameters
    ----------
        yhat : torch.Tensor
            The soft mask from the network of shape (B, C, H, W).
        target : torch.Tensor
            The target matrix of shape (B, H, W)
        activation : str, optional:
            Apply sigmoid or softmax activation before taking argmax.

    Raises
    ------
        ValueError if an illegal activation is given.

    Returns
    -------
        torch.Tensor:
            A confusion matrix tensor of shape (B, num_classes, num_classes).

    """
    yhat_soft = yhat

    allowed = ("sigmoid", "softmax")
    if activation is not None:
        if activation not in allowed:
            raise ValueError(
                f"Illegal activation given. Got '{activation}'. Allowed: {allowed}."
            )

        if activation == "sigmoid":
            yhat_soft = torch.sigmoid(yhat)
        elif activation == "softmax":
            yhat_soft = F.softmax(yhat, dim=1)

    n_classes = yhat_soft.shape[1]
    batch_size = yhat_soft.shape[0]
    bins = target + torch.argmax(yhat_soft, dim=1) * n_classes
    bins_vec = bins.view(batch_size, -1)

    confusion_list = []
    for i in range(batch_size):
        pb = bins_vec[i]
        bin_count = torch.bincount(pb, minlength=n_classes**2)
        confusion_list.append(bin_count)

    confusion_vec = torch.stack(confusion_list)
    confusion_mat = confusion_vec.view(batch_size, n_classes, n_classes).to(
        torch.float32
    )

    return confusion_mat


def iou(
    yhat: torch.Tensor,
    target: torch.Tensor,
    activation: str = None,
    eps: float = 1e-7,
) -> torch.Tensor:
    """Compute the per class intersection over union for dense predictions.

    Parameters
    ----------
        yhat : torch.Tensor
            The soft mask from the network of shape (B, C, H, W).
        target : torch.Tensor
            The target matrix of shape (B, H, W).
        activation : str, optional
            Apply sigmoid or softmax activation before taking argmax.
        eps : float, default=1e-7
            Small constant to avoid zero div error.

    Returns
    -------
        torch.Tensor:
            An iou matrix of shape (B, num_classes, num_classes).
    """
    conf_mat = confusion_mat(yhat, target, activation)
    rowsum = torch.sum(conf_mat, dim=1)  # [(TP + FP), (FN + TN)]
    colsum = torch.sum(conf_mat, dim=2)  # [(TP + FN), (FP + TN)]
    diag = torch.diagonal(conf_mat, dim1=-2, dim2=-1)  # [TP, TN]
    denom = rowsum + colsum - diag  # [(TP + FN + FP), (TN + FN + FP)]
    ious = (diag + eps) / (denom + eps)

    return ious  # [(TP/(TP + FN + FP)), (TN/(TN + FN + FP))]


def accuracy(
    yhat: torch.Tensor,
    target: torch.Tensor,
    activation: str = None,
    eps: float = 1e-7,
) -> torch.Tensor:
    """Compute the per class accuracy for dense predictions.

    Parameters
    ----------
        yhat : torch.Tensor
            The soft mask from the network of shape (B, C, H, W).
        target : torch.Tensor
            The target matrix of shape (B, H, W).
        activation : str, optional
            apply sigmoid or softmax activation before taking argmax
        eps : float, default=1e-7
            Small constant to avoid zero div error.

    Returns
    -------
        torch.Tensor:
            The accuracy. A tensor of shape (1).
    """
    conf_mat = confusion_mat(yhat, target, activation)
    diag = torch.diagonal(conf_mat, dim1=-2, dim2=-1)  # batch diagonal
    denom = conf_mat.sum()
    accuracies = (diag + eps) / (denom + eps)

    return accuracies.sum()



from pathlib import Path
from typing import BinaryIO, Callable, List, Optional, Tuple

import numpy as np
import tables as tb
from tqdm import tqdm

from ._base_writer import BaseWriter

__all__ = ["SlidingWindowHDF5Writer"]


class SlidingWindowHDF5Writer(BaseWriter):
    def __init__(
        self,
        in_dir_im: str,
        in_dir_mask: str,
        save_dir: str,
        file_name: str,
        patch_size: Tuple[int, int],
        stride: int,
        transforms: Optional[List[str]] = None,
        chunk_size: int = 1,
        complevel: int = 5,
        complib: str = "blosc:lz4",
    ) -> None:
        """Write overlapping patches to a hdf5 database.

        Parameters
        ----------
            in_dir_im : str
                Path to the folder of images
            in_dir_mask : str
                Path to the folder of masks.
            save_dir : str
                Path to the folder where the db will be saved.
            file_name : str
                Name of the h5 db-file.
            patch_size : Tuple[int, int]
                Height and width of the extracted patches.
            stride : int
                Stride for the sliding window.
            transforms : List[str]:
                A list of transforms to apply to the images and masks. Allowed ones:
                "blur", "non_spatial", "non_rigid", "rigid", "hue_sat", "random_crop",
                "center_crop", "resize"
            chunk_size : int, default=1
                The chunk size of the h5 arrays.
            complevel : int, default=5
                Specifies a compression level for data. The allowed range is 0-9.
                A value of 0 (the default) disables compression. (PyTables).
            complib : str, default="blosc:lz4"
                Specifies the compression library to be used. 'zlib', 'lzo', 'bzip2' and
                'blosc' are supported. Additional compressors for Blosc: 'blosc:blosclz'
                'blosc:lz4', 'blosc:lz4hc', 'blosc:snappy', 'blosc:zlib' & 'blosc:zstd'.
                (PyTables).

        Raises
        ------
            ValueError if issues with given paths or the files in those folders.

        Example
        -------
            >>> writer = SlidingWindowHDF5Writer(
                    "/path/to/my/imgs/,
                    "/path/to/my/masks/,
                    "/save/db/here/,
                    "my_h5_dataset.h5",
                    patch_size=(320, 320),
                    stride=160,
                    transforms=["rigid"]
                )
            >>> writer.write()
        """
        super().__init__(
            in_dir_im=in_dir_im,
            in_dir_mask=in_dir_mask,
            stride=stride,
            patch_size=patch_size,
            transforms=transforms,
        )

        self.chunk_size = chunk_size
        self.complevel = complevel
        self.complib = complib
        self.save_dir = Path(save_dir)
        self.fname = self.save_dir / file_name
        if self.fname.suffix:
            if self.fname.suffix not in (".h5", ".hdf5"):
                raise ValueError(
                    "`file_name` has to have a '.h5' or '.hdf5' suffix."
                    f"Got: {self.fname.suffix}"
                )
        else:
            self.fname = self.fname.with_suffix(".h5")

    def write(self, pre_proc: Callable = None, msg: str = None) -> None:
        """Write patches to a hdf5 db.

        Parameters
        ----------
            pre_proc : Callable, optional
                An optional pre-processing function for the masks.
        """
        h5 = tb.open_file(self.fname.as_posix(), mode="w")

        h5.create_earray(
            where=h5.root,
            name="imgs",
            atom=tb.UInt8Atom(),
            shape=np.append([0], self.patch_size + (3,)),
            chunkshape=np.append([self.chunk_size], self.patch_size + (3,)),
            filters=tb.Filters(complevel=self.complevel, complib=self.complib),
        )

        h5.create_earray(
            where=h5.root,
            name="insts",
            atom=tb.Int32Atom(),
            shape=np.append([0], self.patch_size),
            chunkshape=np.append([self.chunk_size], self.patch_size),
            filters=tb.Filters(complevel=self.complevel, complib=self.complib),
        )

        h5.create_earray(
            where=h5.root,
            name="types",
            atom=tb.Int32Atom(),
            shape=np.append([0], self.patch_size),
            chunkshape=np.append([self.chunk_size], self.patch_size),
            filters=tb.Filters(complevel=self.complevel, complib=self.complib),
        )

        h5.create_earray(
            where=h5.root,
            name="areas",
            atom=tb.Int32Atom(),
            shape=np.append([0], self.patch_size),
            chunkshape=np.append([self.chunk_size], self.patch_size),
            filters=tb.Filters(complevel=self.complevel, complib=self.complib),
        )

        try:
            with tqdm(
                zip(self.fnames_im, self.fnames_mask), total=len(self.fnames_im)
            ) as pbar:
                total_tiles = 0
                msg = msg if msg is not None else ""
                for fni, fnm in pbar:
                    pbar.set_description(f"Extracting {msg} patches to h5 db..")
                    tiles = self._get_tiles(fni, fnm, pre_proc)
                    n_tiles = tiles["image"].shape[0]

                    arg_list = []
                    for i in range(n_tiles):
                        dd = dict(h5=h5)
                        for k, m in tiles.items():
                            dd[k] = m[i : i + 1]
                        arg_list.append(dd)
                    self._write_sequential(self.save2db, arg_list)
                    total_tiles += n_tiles
                    pbar.set_postfix_str(f"# of extracted tiles {total_tiles}")
        except Exception as e:
            h5.close()
            print(e)

        h5.root._v_attrs.n_items = h5.root.imgs.shape[0]
        h5.close()

    def save2db(
        self,
        h5: BinaryIO,
        image: np.ndarray,
        inst_map: np.ndarray,
        type_map: np.ndarray = None,
        sem_map: np.ndarray = None,
    ) -> None:
        """Write image and corresponding masks to h5-db."""
        h5.root.imgs.append(image)
        h5.root.insts.append(inst_map)

        if type_map is not None:
            h5.root.types.append(type_map)
        if sem_map is not None:
            h5.root.areas.append(sem_map)



from typing import Tuple

import torch
import torch.nn as nn

from ...modules import Identity
from .unet import UnetSkip
from .unet3p import Unet3pSkip
from .unetpp import UnetppSkip

__all__ = ["LongSkip"]


LONGSKIP_LOOKUP = {
    "unet": UnetSkip,
    "unet3p": Unet3pSkip,
    "unet3p-lite": Unet3pSkip,
    "unetpp": UnetppSkip,
}


class LongSkip(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Long skip wrapper class.

        Parameters
        ----------
            name : str
                The name of the long skip method.
                One of "unet", "unet3p", "unet3p-lite", "unetpp".

        Raises
        ------
            ValueError: if the long skip method name is illegal.
        """
        super().__init__()

        allowed = list(LONGSKIP_LOOKUP.keys()) + [None]
        if name not in allowed:
            raise ValueError(
                f"Illegal long skip method given. Allowed: {allowed}. Got: '{name}'"
            )

        if name is not None:
            if name == "unet3p-lite":
                kwargs["lite_version"] = True
            self.skip = LONGSKIP_LOOKUP[name](**kwargs)
        else:
            self.skip = Identity()

    @property
    def out_channels(self) -> int:
        """Out channels."""
        return self.skip.out_channels

    def forward(
        self, x: torch.Tensor, skips: Tuple[torch.Tensor, ...], **kwargs
    ) -> torch.Tensor:
        """Forward pass of the long skip function."""
        return self.skip(x, skips, **kwargs)



import pytest

from cellseg_models_pytorch.models import cellpose_base
from cellseg_models_pytorch.optimizers import OPTIM_LOOKUP, adjust_optim_params


@pytest.mark.parametrize("optim", ["adam"])
def test_optim_setup(optim):
    model = cellpose_base(type_classes=3)
    optim_params = {
        "encoder": {"lr": 0.004, "weight_decay": 0.3},
        "decoder": {"lr": 0.003, "weight_decay": 0.2},
    }
    params = adjust_optim_params(model, optim_params)
    optimizer = OPTIM_LOOKUP[optim](params)

    assert all([p_g["lr"] == p["lr"] for p_g, p in zip(optimizer.param_groups, params)])
    assert all(
        [
            p_g["weight_decay"] == p["weight_decay"]
            for p_g, p in zip(optimizer.param_groups, params)
        ]
    )



"""
Imported from cellpose repo (Old version). Code style mods.

https://github.com/MouseLand/cellpose/blob/master/cellpose/dynamics.py

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from typing import Union

import numpy as np
import torch
from scipy.ndimage.filters import maximum_filter1d

from cellseg_models_pytorch.utils import binarize

from .utils import fill_holes_and_remove_small_masks

__all__ = ["post_proc_cellpose_old"]


def steps2D_interp_torch(p: np.ndarray, dP: np.ndarray, niter: int = 200) -> np.ndarray:
    """Run dynamics.

    "run a dynamical system starting at that pixel location and
     following the spatial derivatives specified by the horizontal and
     vertical gradient maps. We use finite differences with a step size
     of one."

    Parameters
    ----------
        p : float32, 3D array
            final locations of each pixel after dynamics,
            size [axis x Ly x Lx].
        dP : float32, 3D array
            flows [axis x Ly x Lx]
        niter : int (optional, default 200)
            number of iterations of dynamics to run

    Returns
    -------
        np.ndarray:
            The pixel locations after dynamics. Shape (2, H, W).
    """
    shape = dP.shape[1:]
    device = "cuda" if torch.cuda.is_available() else "cpu"

    pt = torch.from_numpy(p[[1, 0]].T).double().to(device)
    pt = pt.unsqueeze(0).unsqueeze(0)
    pt[..., 0] = pt[..., 0] / (shape[1] - 1)  # normalize to between  0 and 1
    pt[..., 1] = pt[..., 1] / (shape[0] - 1)  # normalize to between  0 and 1
    pt = pt * 2 - 1  # normalize to between -1 and 1
    im = torch.from_numpy(dP[[1, 0]]).double().to(device)
    im = im.unsqueeze(0)

    for k in range(2):
        im[:, k, ...] /= (shape[1 - k] - 1) / 2.0

    for _ in range(niter):
        dPt = torch.nn.functional.grid_sample(im, pt, align_corners=False)
        for k in range(2):
            pt[..., k] = torch.clamp(pt[..., k] - dPt[:, k, ...], -1.0, 1.0)

    pt = (pt + 1) * 0.5
    pt[..., 0] = pt[..., 0] * (shape[1] - 1)
    pt[..., 1] = pt[..., 1] * (shape[0] - 1)

    return pt[..., [1, 0]].cpu().numpy().squeeze().T


def follow_flows(dP: np.ndarray, niter: int = 200) -> np.ndarray:
    """Define pixels and run dynamics to recover masks in 2D.

    Pixels are meshgrid. Only pixels with non-zero cell-probability
    are used (as defined by inds)

    Parameters
    ----------
        dP : float32, 3D
            flows [axis x Ly x Lx]
        niter : int (optional, default 200)
            number of iterations of dynamics to run

    Returns
    -------
        p : float32, 3D array
            Final locations of each pixel after dynamics
    """
    shape = np.array(dP.shape[1:]).astype(np.int32)
    niter = np.int32(niter)

    p = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing="ij")

    p = np.array(p).astype(np.float32)

    # run dynamics on subset of pixels. NOTE: Uses only xmap or ymap (dP[0|1])
    inds = np.array(np.nonzero(np.abs(dP[1]) > 1e-3)).astype(np.int32).T

    # Interpolation mode
    # Sometimes a random error in empty images.. errr... dunno...
    try:
        p[:, inds[:, 0], inds[:, 1]] = steps2D_interp_torch(
            p=p[:, inds[:, 0], inds[:, 1]],
            dP=dP,
            niter=niter,
        )
    except Exception:
        pass

    return p


def get_masks_cellpose_old(
    p: np.ndarray,
    threshold: int = 0.4,
    rpad: int = 20,
    flows: Union[np.ndarray, None] = None,
    iscell: Union[np.ndarray, None] = None,
) -> np.ndarray:
    """Create masks using pixel convergence after running dynamics.

    Makes a histogram of final pixel locations p, initializes masks
    at peaks of histogram and extends the masks from the peaks so that
    they include all pixels with more than 2 final pixels p. Discards
    masks with flow errors greater than the threshold.

    Parameters
    ----------
        p: float32, 3D array
            final locations of each pixel after dynamics,
            size [axis x Ly x Lx].
        rpad: int (optional, default 20)
            histogram edge padding
        threshold: float (optional, default 0.4)
            masks with flow error greater than threshold are discarded
            (if flows is not None)
        flows: float, 3D array (optional, default None)
            flows [axis x Ly x Lx]. If flows
            is not None, then masks with inconsistent flows are removed using
            `remove_bad_flow_masks`.
        iscell: bool, 2D array
            if iscell is not None, set pixels that are
            iscell False to stay in their original location.
    Returns
    -------
        M0: int, 2D array
            masks with inconsistent flow masks removed,
            0=NO masks; 1,2,...=mask labels,
            size [Ly x Lx]

    """
    pflows = []
    edges = []
    shape0 = p.shape[1:]
    dims = len(p)
    if iscell is not None:
        inds = np.meshgrid(np.arange(shape0[0]), np.arange(shape0[1]), indexing="ij")

        for i in range(dims):
            p[i, ~iscell] = inds[i][~iscell]

    for i in range(dims):
        pflows.append(p[i].flatten().astype("int32"))
        edges.append(np.arange(-0.5 - rpad, shape0[i] + 0.5 + rpad, 1))

    h, _ = np.histogramdd(tuple(pflows), bins=edges)
    hmax = h.copy()
    for i in range(dims):
        hmax = maximum_filter1d(hmax, 5, axis=i)

    seeds = np.nonzero(np.logical_and(h - hmax > -1e-6, h > 10))
    Nmax = h[seeds]
    isort = np.argsort(Nmax)[::-1]
    for s in seeds:
        s = s[isort]

    pix = list(np.array(seeds).T)
    shape = h.shape
    expand = np.nonzero(np.ones((3, 3)))
    for e in expand:
        e = np.expand_dims(e, 1)

    for iter in range(5):
        for k in range(len(pix)):
            if iter == 0:
                pix[k] = list(pix[k])
            newpix = []
            iin = []
            for i, e in enumerate(expand):
                epix = e[:, None] + np.expand_dims(pix[k][i], 0) - 1
                epix = epix.flatten()
                iin.append(np.logical_and(epix >= 0, epix < shape[i]))
                newpix.append(epix)
            iin = np.all(tuple(iin), axis=0)
            for p in newpix:
                p = p[iin]
            newpix = tuple(newpix)
            igood = h[newpix] > 2
            for i in range(dims):
                pix[k][i] = newpix[i][igood]
            if iter == 4:
                pix[k] = tuple(pix[k])

    M = np.zeros(h.shape, np.int32)
    for k in range(len(pix)):
        M[pix[k]] = 1 + k

    for i in range(dims):
        pflows[i] = pflows[i] + rpad

    # remove big masks
    M0 = M[tuple(pflows)]
    _, counts = np.unique(M0, return_counts=True)
    big = np.prod(shape0) * 1.0
    for i in np.nonzero(counts > big)[0]:
        M0[M0 == i] = 0

    _, M0 = np.unique(M0, return_inverse=True)
    M0 = np.reshape(M0, shape0)

    return M0


def post_proc_cellpose_old(inst_map: np.ndarray, flow_map: np.ndarray) -> np.ndarray:
    """Run the CellPose post-processing pipeline (OLD).

    https://www.nature.com/articles/s41592-020-01018-x

    Note:
    Does not discard bad flows which are removed in the original paper.
    In general the flows are somewhat useless... except for nice viz..

    Args:
    -----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        flow_map : np.ndarray
            Horizontal and vertical flows. Shape (2, H, W).

    Returns
    -------
        np.ndarray:
            The instance labelled segmentation mask. Shape (H, W)

    """
    binary_mask = binarize(inst_map).astype(bool)
    # dP = -1 * dp * binary_mask / 5.0 # Weird result, Dunno ??
    dP = flow_map * binary_mask
    pixel_loc = follow_flows(dP, niter=3000)

    mask = get_masks_cellpose_old(
        p=pixel_loc, iscell=binary_mask, flows=dP, threshold=None
    )
    inst_map = fill_holes_and_remove_small_masks(mask, min_size=30)

    return inst_map.astype("i4")



from timeit import repeat
from typing import Any, Callable, Dict, List, Tuple, Union

import numpy as np
import torch
from tqdm import tqdm

from cellseg_models_pytorch.inference import BaseInferer

__all__ = ["LatencyBenchmarker"]


class LatencyBenchmarker:
    def __init__(self, inferer: BaseInferer) -> None:
        """Benchmark latencies of the model an post-processing pipelines."""
        self.inferer = inferer

        try:
            self.inferer.out_masks
            self.inferer.soft_masks
        except AttributeError:
            raise AttributeError(
                "Did not find `out_masks` or `soft_masks` attributes. "
                "To get these, run inference with `inferer.infer()`. "
                "Remember to set the `save_intermediate param to True for the inferer.`"
            )

    def inference_latency(
        self, reps: int = 1, warmup_reps: int = 1, **kwargs
    ) -> List[Dict[str, Any]]:
        """Compute the inference-pipeline latency.

        Parameters
        ----------
            reps : int, defalt=1
                Repetition per batch.
            warmup_reps : int, default=1
                Warm up repetitions.

        Returns
        -------
            List[Dict[str, Any]]:
                Latency info info of the samples.
        """
        times = []
        with tqdm(self.inferer.dataloader, unit="batch") as loader:
            with torch.no_grad():
                for data in loader:
                    res = self._compute_latency(
                        self.inferer._infer_batch,
                        maps={"input_batch": data["im"]},
                        reps=reps,
                        warmup_reps=warmup_reps,
                        **kwargs,
                    )
                    bsize = data["im"].shape[0]
                    res["n_images"] = bsize
                    res["input_shape"] = tuple(data["im"].shape[1:])
                    res["throughput (img/s)"] = res["mean latency(s)"] / bsize
                    times.append(res)

        return times

    def inference_postproc_latency(self, reps: int = 1) -> List[Dict[str, Any]]:
        """Compute the latency of the whole inference + post-processing pipeline.

        Parameters
        ----------
            reps : int, default=1
                Number of repetitions of the full pipeline.

        Returns
        -------
            List[Dict[str, Any]]
                The latency and throughput info of the pipeline.
        """
        timings = repeat(
            lambda: self.inferer.infer(),
            repeat=reps,
            number=1,
        )

        timings = np.array(timings)
        mean_syn = np.mean(timings)
        std_syn = np.std(timings)

        res = {}
        res["n_images"] = len(list(self.inferer.soft_masks.keys()))
        res["repetitions"] = reps
        res["total mean latency(s)"] = mean_syn
        res["total std latency(s)"] = std_syn
        res["mean throughput (img/s)"] = mean_syn / res["n_images"]
        res["std throughput (img/s))"] = std_syn / res["n_images"]

        return [res]

    def postproc_latency(
        self, which: str = "inst", reps_per_img: int = 10, **kwargs
    ) -> List[Dict[str, Any]]:
        """Compute the post-processing latencies.

        Parameters
        ----------
            which : str, default="inst"
                Which post-processing type. One of "inst", "sem", "type".
            reps_per_img : int, default=10
                Number of repetitions per image.
            **kwargs:
                Arbitrary keyword args for the post-proc func.

        Returns
        -------
            List[Dict[str, Any]]:
                A list of dicts containing throughput info of each of the samples.
        """
        PP_LOOKUP = {
            "sem": self._compute_sem_postproc_latency,
            "type": self._compute_type_postproc_latency,
            "inst": self._compute_inst_postproc_latency,
        }

        allowed = list(PP_LOOKUP.keys())
        if which not in allowed:
            raise ValueError(f"Illegal `type` arg. Got: {type}. Allowed: {allowed}")

        aux_key = self.inferer.postprocessor.aux_key
        inst_key = self.inferer.postprocessor.inst_key
        samples = list(self.inferer.soft_masks.keys())

        times = []
        with tqdm(samples, total=len(samples)) as pbar:
            for k in samples:
                if which == "inst":
                    res = PP_LOOKUP["inst"](
                        prob_map=self.inferer.soft_masks[k][inst_key],
                        aux_map=self.inferer.soft_masks[k][aux_key],
                        return_cell_count=True,
                        reps=reps_per_img,
                        **kwargs,
                    )
                elif which == "type":
                    res = PP_LOOKUP["type"](
                        prob_map=self.inferer.soft_masks[k][inst_key],
                        inst_map=self.inferer.out_masks[k]["inst"],
                        reps=reps_per_img,
                        **kwargs,
                    )
                elif which == "sem":
                    res = PP_LOOKUP["sem"](
                        prob_map=self.inferer.soft_masks[k]["sem"],
                        reps=reps_per_img,
                        **kwargs,
                    )

                res["name"] = k
                times.append(res)
                pbar.update(1)

        return times

    def model_latency(
        self,
        input_size: Tuple[int, int] = (256, 256),
        batch_size: int = 1,
        reps: int = 100,
        warmup_reps: int = 3,
        device="cuda",
    ) -> List[Dict[str, Any]]:
        """Measure the model inference latency in secods.

        I.e. one forward pass of the model.

        Parameters
        ----------
            input_size : Tuple[int, int]
                Height and width of the input.
            batch_size : int, default=1
                Batch size.
            reps : int, default=300
                Number of repetitions to run the latency measurement.
            warmup_reps : int, default=3
                Number of repetitions that are used for warming up the gpu.
                I.e. the number of repetitions that are excluded from the
                beginning.
            device : str, default="cuda"
                One of 'cuda' or 'cpu'.

        Returns
        -------
            List[Dict[str, Any]]:
                The latency mean and standard deviation in secods + extra info.
        """
        dummy_input = torch.randn(
            batch_size,
            3,
            input_size[0],
            input_size[1],
            dtype=torch.float,
            device=device,
        )

        if device == "cpu":
            self.inferer.predictor.model.cpu()

        starter = torch.cuda.Event(enable_timing=True)
        ender = torch.cuda.Event(enable_timing=True)

        timings = np.zeros((reps, 1))
        with tqdm(total=reps, unit="rep") as pbar:
            with torch.no_grad():
                for rep in range(reps):
                    starter.record()
                    _ = self.inferer.predictor.forward_pass(dummy_input)
                    ender.record()

                    # wait for gpu sync
                    torch.cuda.synchronize()
                    curr_time = starter.elapsed_time(ender) / 1000
                    timings[rep] = curr_time
                    pbar.update(1)

        mean_syn = np.sum(timings[warmup_reps:]) / (reps - warmup_reps)
        std_syn = np.std(timings[warmup_reps:])
        shape = tuple(dummy_input.shape[1:])

        res = {
            "batch_size": batch_size,
            "input_shape": shape,
            "mean latency(s)": mean_syn,
            "std latency(s)": std_syn,
        }

        return [res]

    def model_throughput(
        self,
        input_size: Tuple[int, int] = (256, 256),
        batch_size: int = 1,
        reps: int = 100,
        warmup_reps: int = 3,
        device="cuda",
    ) -> List[Dict[str, Any]]:
        """Measure the inference throughput in seconds.

        I.e. Measure model forward pass throughput (image/s).

        Parameters
        ----------
            input_size : Tuple[int, int]
                Height and width of the input.
            batch_size : int, default=1
                Batch size for the model.
            reps : int, default=300
                Number of repetitions to run the latency measurement.
            warmup_reps : int, default=3
                Number of repetitions that are used for warming up the gpu.
                I.e. the number of repetitions that are excluded from the
                beginning.
            device : str, default="cuda"
                One of 'cuda' or 'cpu'.

        Returns
        -------
            List[Dict[str, Any]]:
                The throughput of the model (image/s) + extra info.
        """
        dummy_input = torch.randn(
            batch_size,
            3,
            input_size[0],
            input_size[1],
            dtype=torch.float,
            device=device,
        )

        if device == "cpu":
            self.inferer.predictor.model.cpu()

        total_time = 0

        with torch.no_grad():
            for _ in range(reps):
                starter = torch.cuda.Event(enable_timing=True)
                ender = torch.cuda.Event(enable_timing=True)
                starter.record()
                _ = self.inferer.predictor.forward_pass(dummy_input)
                ender.record()
                torch.cuda.synchronize()
                curr_time = starter.elapsed_time(ender) / 1000
                total_time += curr_time

        throughput = ((reps - warmup_reps) * batch_size) / total_time
        shape = tuple(dummy_input.shape[1:])

        res = {
            "batch_size": batch_size,
            "input_shape": shape,
            "throughput(img/s)": throughput,
        }

        return [res]

    def _compute_latency(
        self,
        func: Callable,
        maps: Dict[str, Union[np.ndarray, torch.Tensor]],
        reps: int = 300,
        warmup_reps: int = 3,
        **kwargs,
    ) -> Dict[str, Any]:
        """Run the latency measurements."""
        if kwargs:
            kwargs = {**maps, **kwargs}
        else:
            kwargs = maps

        timings = repeat(
            lambda: func(**kwargs),
            repeat=reps + warmup_reps,
            number=1,
        )
        timings = np.array(timings)

        mean_syn = np.mean(timings[warmup_reps:])
        std_syn = np.std(timings[warmup_reps:])

        res = {}
        res["repetitions"] = reps
        res["mean latency(s)"] = mean_syn
        res["std latency(s)"] = std_syn

        return res

    def _compute_inst_postproc_latency(
        self,
        prob_map: np.ndarray,
        aux_map: np.ndarray,
        reps: int = 10,
        warmup_reps: int = 2,
        **kwargs,
    ) -> Dict[str, Any]:
        """Compute the instance segmentation post-proc latency in seconds.

        I.e. One run of the post-procesing method.

        NOTE: returns also additional data that affects the latency such
        as the number of objects and number of object pixels in the image.

        Parameters
        ----------
            prob_map : np.ndarray
                The probability map of the object instances. Shape: (C, H, W).
            aux_map : np.ndarray
                The auxilliary regression output. Shape: (C, H, W).
            reps : int, default=10
                Number of repetitions for timeit.
            warmup_reps : int, default=2
                Warmup loops for the function.
            **kwargs:
                Arbitrary keyword args for the post-proc func.

        Returns
        -------
            Dict[str, Any]:
                A dictionary with data related to the sample latency.
        """
        res = self._compute_latency(
            self.inferer.postprocessor._get_inst_map,
            maps={"prob_map": prob_map, "aux_map": aux_map},
            reps=reps,
            warmup_reps=warmup_reps,
            **kwargs,
        )

        x = self.inferer.postprocessor._get_inst_map(prob_map, aux_map, **kwargs)
        cells, counts = np.unique(x, return_counts=True)
        res["input_shape"] = x.shape
        res["ncells"] = len(cells)
        res["npixels"] = np.sum(counts[1:])

        return res

    def _compute_sem_postproc_latency(
        self, prob_map: np.ndarray, reps: int = 10, warmup_reps: int = 2, **kwargs
    ) -> List[Dict[str, Any]]:
        """Compute the semantic segmentation post-proc latency in seconds.

        I.e. One run of the post-procesing method.

        Parameters
        ----------
            prob_map : np.ndarray
                The probability map of the semantic segmentation. Shape: (C, H, W).
            reps : int, default=10
                Number of repetitions for timeit.
            warmup_reps : int, default=2
                Warmup loops for the function.
            **kwargs:
                Arbitrary keyword args for the post-proc func.

        Returns
        -------
            Dict[str, Any]:
                A dictionary with data related to the sample latency.
        """
        res = self._compute_latency(
            self.inferer.postprocessor._get_sem_map,
            maps={"prob_map": prob_map},
            reps=reps,
            warmup_reps=warmup_reps,
            **kwargs,
        )

        x = self.inferer.postprocessor._get_sem_map(prob_map, **kwargs)
        _, npixels = np.unique(x, return_counts=True)
        res["input_shape"] = x.shape
        res["npixels"] = np.sum(npixels[1:])

        return res

    def _compute_type_postproc_latency(
        self,
        prob_map: np.ndarray,
        inst_map: np.ndarray,
        reps: int = 10,
        warmup_reps: int = 2,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """Compute the type segmentation post-proc latency in seconds.

        I.e. One run of the post-procesing method.

        Parameters
        ----------
            prob_map : np.ndarray
                The probability map of the object instances. Shape: (C, H, W).
            inst_map : np.ndarray
                The labelled instance segmentation map. Shape: (H, W).
            reps : int, default=10
                Number of repetitions for timeit.
            warmup_reps : int, default=2
                Warmup loops for the function.
            **kwargs:
                Arbitrary keyword args for the post-proc func.

        Returns
        -------
            Dict[str, Any]:
                A dictionary with data related to the sample latency.
        """
        res = self._compute_latency(
            self.inferer.postprocessor._get_type_map,
            maps={"prob_map": prob_map, "inst_map": inst_map},
            reps=reps,
            warmup_reps=warmup_reps,
            **kwargs,
        )

        x = self.inferer.postprocessor._get_type_map(
            prob_map, inst_map, use_mask=True, **kwargs
        )
        cells = np.unique(inst_map)
        _, npixels = np.unique(x, return_counts=True)
        res["img_shape"] = x.shape
        res["ncells"] = len(cells)
        res["npixels"] = np.sum(npixels[1:])

        return res



from abc import ABC, abstractmethod
from collections import OrderedDict
from itertools import chain
from pathlib import Path
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from pathos.multiprocessing import ThreadPool as Pool
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..utils import tensor_to_ndarray
from ..utils.save_utils import mask2mat
from .folder_dataset import FolderDataset
from .post_processor import PostProcessor
from .predictor import Predictor


class BaseInferer(ABC):
    def __init__(
        self,
        model: nn.Module,
        input_folder: Union[Path, str],
        out_activations: Dict[str, str],
        out_boundary_weights: Dict[str, bool],
        patch_size: Tuple[int, int],
        instance_postproc: str,
        padding: int = None,
        batch_size: int = 8,
        normalization: str = None,
        device: str = "cuda",
        save_masks: bool = True,
        save_intermediate: bool = False,
        save_dir: Union[Path, str] = None,
        checkpoint_path: Union[Path, str] = None,
        n_images: int = None,
        type_post_proc: Callable = None,
        sem_post_proc: Callable = None,
        **postproc_kwargs,
    ) -> None:
        """Inference for an image folder.

        Parameters
        ----------
            model : nn.Module
                A segmentation model.
            input_folder : Path | str
                Path to a folder of images.
            out_activations : Dict[str, str]
                Dictionary of head names mapped to a string value that specifies the
                activation applied at the head. E.g. {"type": "tanh", "cellpose": None}
                Allowed values: "softmax", "sigmoid", "tanh", None.
            out_boundary_weights : Dict[str, bool]
                Dictionary of head names mapped to a boolean value. If the value is
                True, after a prediction, a weight matrix is applied that assigns bigger
                weight on pixels in the center and less weight to pixels on the tile
                boundaries. helps dealing with prediction artefacts on the boundaries.
                E.g. {"type": False, "cellpose": True}
            patch_size : Tuple[int, int]:
                The size of the input patches that are fed to the segmentation model.
            instance_postproc : str
                The post-processing method for the instance segmentation mask. One of:
                "cellpose", "omnipose", "stardist", "hovernet", "dcan", "drfns", "dran"
            padding : int, optional
                The amount of reflection padding for the input images.
            batch_size : int, default=8
                Number of images loaded from the folder at every batch.
            normalization : str, optional
                Apply img normalization at forward pass (Same as during training).
                One of: "dataset", "minmax", "norm", "percentile", None.
            device : str, default="cuda"
                The device of the input and model. One of: "cuda", "cpu"
            save_masks : bool, default=False
                If True, the resulting segmentation masks will be saved into `out_masks`
                variable.
            save_intermediate : bool, default=False
                If True, intermediate soft masks will be saved into `soft_masks` var.
            save_dir : bool, optional
                Path to save directory. If None, no masks will be saved to disk as .mat
                files. If not None, overrides `save_masks`, thus for every batch the
                segmentation results are saved into disk and the intermediate results
                are flushed.
            checkpoint_path : Path | str, optional
                Path to the model weight checkpoints.
            n_images : int, optional
                First n-number of images used from the `input_folder`.
            type_post_proc : Callable, optional
                A post-processing function for the type maps. If not None, overrides
                the default.
            sem_post_proc : Callable, optional
                A post-processing function for the semantc seg maps. If not None,
                overrides the default.
            **postproc_kwargs:
                Arbitrary keyword arguments for the post-processing.
        """
        self.save_dir = Path(save_dir) if save_dir is not None else None
        self.save_masks = save_masks
        self.save_intermediate = save_intermediate

        # dataloader
        self.path = Path(input_folder)
        folder_ds = FolderDataset(self.path, n_images=n_images)
        self.dataloader = DataLoader(
            folder_ds, batch_size=batch_size, shuffle=False, pin_memory=True
        )

        # model and device
        self.model = model
        if device == "cpu":
            self.model.cpu()
            self.device = torch.device("cpu")
        if torch.cuda.is_available() and device == "cuda":
            self.model.cuda()
            self.device = torch.device("cuda")

        self.model.eval()

        if checkpoint_path is not None:
            ckpt = torch.load(
                checkpoint_path, map_location=lambda storage, loc: storage
            )

            try:
                self.model.load_state_dict(ckpt["state_dict"], strict=True)
            except RuntimeError:
                new_ckpt = self._strip_state_dict(ckpt)
                self.model.load_state_dict(new_ckpt["state_dict"], strict=True)
            except BaseException as e:
                print(e)

        #
        self.predictor = Predictor(
            model=self.model,
            patch_size=patch_size,
            normalization=normalization,
            device=self.device,
        )
        self.out_heads = self._get_out_info()  # the names and num channels of out heads
        self.batch_size = batch_size
        self.patch_size = patch_size
        self.padding = padding
        self.out_activations = out_activations
        self.out_boundary_weights = out_boundary_weights
        self.head_kwargs = self._check_and_set_head_args()

        #
        self.postprocessor = PostProcessor(
            instance_postproc,
            inst_key=self.model.inst_key,
            aux_key=self.model.aux_key,
            type_post_proc=type_post_proc,
            sem_post_proc=sem_post_proc,
            **postproc_kwargs,
        )

    @abstractmethod
    def _infer_batch(self):
        raise NotImplementedError

    def infer(self) -> None:
        """Run inference and post-processing for the images.

        NOTE: Saves outputs in `self.out_masks` or to disk (.mat) files.

        `self.out_masks` is a nested dict: E.g.
            {"image1": {"inst": [H, W], "type": [H, W], "sem": [H, W]}}
        """
        self.soft_masks = {}
        self.out_masks = {}
        self.elapsed = []
        self.rate = []
        with tqdm(self.dataloader, unit="batch") as loader:
            with torch.no_grad():
                for data in loader:
                    names = data["file"]
                    loader.set_description("Running inference")
                    loader.set_postfix_str("Forward pass")
                    soft_masks = self._infer_batch(data["im"])
                    loader.set_postfix_str("post-processing")
                    soft_masks = self._prepare_mask_list(names, soft_masks)

                    # use multi-threading if batch size more than 1.
                    if self.batch_size > 1:
                        seg_results = self.postprocessor.run_parallel(soft_masks)
                    else:
                        seg_results = [
                            self.postprocessor.post_proc_pipeline(soft_masks[0])
                        ]

                    self.elapsed.append(loader.format_dict["elapsed"])
                    self.rate.append(loader.format_dict["rate"])

                    if self.save_intermediate:
                        for n, m in zip(names, soft_masks):
                            self.soft_masks[n] = m

                    if self.save_dir is None:
                        if self.save_masks:
                            for n, m in zip(names, seg_results):
                                self.out_masks[n] = m
                    else:
                        loader.set_postfix_str("Saving results to disk")
                        if self.batch_size > 1:
                            self.save_parallel(seg_results, names, self.save_dir)
                        else:
                            for n, m in zip(names, seg_results):
                                self.save_mask(m, n, self.save_dir)

    @staticmethod
    def save_mask(
        maps: Dict[str, np.ndarray],
        fname: str,
        save_dir: Union[str, Path],
        format: str = ".mat",
    ) -> None:
        """Save model outputs to .mat or geojson.

        Parameters
        ----------
            maps : Dict[str, np.ndarray]
                model output names mapped to model outputs.
                E.g. {"sem": np.ndarray, "type": np.ndarray, "inst": np.ndarray}
            fname : str
                Name for the output-file.
            save_dir : Path or str
                Path to the save directory.
            format : str
                One of ".mat" or "geojson"
        """
        allowed = (".mat", ".json")
        if format not in allowed:
            raise ValueError(
                f"Illegal file-format. Got: {format}. Allowed formats: {allowed}"
            )

        if format == ".mat":
            mask2mat(fname, save_dir, **maps)
        else:
            pass

        return True

    @staticmethod
    def save_parallel(
        maps: List[Dict[str, np.ndarray]],
        fnames: List[str],
        save_dir: Union[Path, str],
        format: str = ".mat",
        progress_bar: bool = False,
    ) -> None:
        """Save the model output masks to a folder. (multi-threaded).

        Parameters
        ----------
            maps : List[Dict[str, np.ndarray]]
                The model output map dictionaries in a list.
            fnames : List[str]
                Name for the output-files. (In the same order with `maps`)
            save_dir : Path or str
                Path to the save directory.
            format : str
                One of ".mat" or "geojson"
            progress_bar : bool, default=False
                If True, a tqdm progress bar is shown.
        """
        args = tuple(zip(maps, fnames, [save_dir] * len(maps), [format] * len(maps)))

        with Pool() as pool:
            if progress_bar:
                it = tqdm(pool.imap(BaseInferer._save_mask, args), total=len(maps))
            else:
                it = pool.imap(BaseInferer._save_mask, args)

            for _ in it:
                pass

    @staticmethod
    def _save_mask(args: Tuple[Dict[str, np.ndarray], str, str]) -> None:
        """Unpacks the args for `save_mask` to enable multi-threading."""
        return BaseInferer.save_mask(*args)

    def _strip_state_dict(self, ckpt: Dict) -> OrderedDict:
        """Strip te first 'model.' (generated by lightning) from the state dict keys."""
        state_dict = OrderedDict()
        for k, w in ckpt["state_dict"].items():
            if "num_batches_track" not in k:
                new_key = k.strip("model")[1:]
                state_dict[new_key] = w
        ckpt["state_dict"] = state_dict

        return ckpt

    def _check_and_set_head_args(self) -> None:
        """Check the model has matching head names with the head args and set them."""
        heads = [head_name for head_name, _ in self.out_heads]
        if self.out_boundary_weights.keys() != self.out_activations.keys():
            raise ValueError(
                "Mismatching head names in `out_boundary_weights` & `out_activations`",
            )

        pred_kwargs = {}
        for head_name, val in self.out_activations.items():
            if head_name in heads:
                pred_kwargs[head_name] = {"act": val}
                pred_kwargs[head_name]["apply_weights"] = self.out_boundary_weights[
                    head_name
                ]
            else:
                raise ValueError(
                    f"Mismatching head name. The model contains heads: {heads} ",
                    f"Got: {head_name}",
                )

        return pred_kwargs

    def _prepare_mask_list(
        self, fnames: List[str], soft_masks: torch.Tensor
    ) -> List[Dict[str, np.ndarray]]:
        """Convert the model outputs into arguments for parallel post-processing."""
        softies = {}
        for k, soft_mask in soft_masks.items():
            softies[k] = dict(zip(fnames, tensor_to_ndarray(soft_mask, "BCHW", "BCHW")))

        args = []
        for fn in fnames:
            map_dict = {}
            for mn, d in softies.items():
                map_dict[mn] = d[fn]
            args.append(map_dict)

        return args

    def _get_out_info(self) -> Tuple[Tuple[str, int]]:
        """Get the output names and number of out channels."""
        return tuple(
            chain.from_iterable(
                list(self.model.heads[k].items()) for k in self.model.heads.keys()
            )
        )



"""
Imported from cellpose repo. Code style mods.

https://github.com/MouseLand/cellpose/blob/master/cellpose/dynamics.py

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
from scipy.ndimage.filters import maximum_filter1d
from skimage.filters import apply_hysteresis_threshold

from cellseg_models_pytorch.utils import binarize

from .integrator import follow_flows
from .utils import fill_holes_and_remove_small_masks, gen_flows

__all__ = ["post_proc_cellpose", "get_masks_cellpose"]


def get_masks_cellpose(
    p: np.ndarray,
    mask: np.ndarray,
    rpad: int = 20,
) -> np.ndarray:
    """Create masks using pixel convergence after running dynamics.

    Makes a histogram of final pixel locations p, initializes masks
    at peaks of histogram and extends the masks from the peaks so that
    they include all pixels with more than 2 final pixels p. Discards
    masks with flow errors greater than the threshold.

    Parameters
    ----------
        p : np.ndarray
            Final locations of each pixel after dynamics. Shape (2, H, W).
        mask : np.ndarray
            The binary mask of the cells. Shape (H, W).
        rpad : int, default=20
            Histogram edge padding.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape (H, W).
    """
    shape0 = p.shape[1:]
    dims = len(p)

    inds = np.meshgrid(np.arange(shape0[0]), np.arange(shape0[1]), indexing="ij")

    for i in range(dims):
        p[i, ~mask] = inds[i][~mask]

    pflows = []
    edges = []
    for i in range(dims):
        pflows.append(p[i].flatten().astype("int32"))
        edges.append(np.arange(-0.5 - rpad, shape0[i] + 0.5 + rpad, 1))

    h, _ = np.histogramdd(tuple(pflows), bins=edges)
    hmax = h.copy()
    for i in range(dims):
        hmax = maximum_filter1d(hmax, 5, axis=i)

    seeds = np.nonzero(np.logical_and(h - hmax > -1e-6, h > 10))
    Nmax = h[seeds]
    isort = np.argsort(Nmax)[::-1]
    for s in seeds:
        s = s[isort]

    pix = list(np.array(seeds).T)
    shape = h.shape
    expand = np.nonzero(np.ones((3, 3)))
    for e in expand:
        e = np.expand_dims(e, 1)

    for iter in range(5):
        for k in range(len(pix)):
            if iter == 0:
                pix[k] = list(pix[k])
            newpix = []
            iin = []
            for i, e in enumerate(expand):
                epix = e[:, None] + np.expand_dims(pix[k][i], 0) - 1
                epix = epix.flatten()
                iin.append(np.logical_and(epix >= 0, epix < shape[i]))
                newpix.append(epix)
            iin = np.all(tuple(iin), axis=0)
            for p in newpix:
                p = p[iin]
            newpix = tuple(newpix)
            igood = h[newpix] > 2
            for i in range(dims):
                pix[k][i] = newpix[i][igood]
            if iter == 4:
                pix[k] = tuple(pix[k])

    M = np.zeros(h.shape, np.int32)
    for k in range(len(pix)):
        M[pix[k]] = 1 + k

    for i in range(dims):
        pflows[i] = pflows[i] + rpad

    # remove big masks
    M0 = M[tuple(pflows)]
    _, counts = np.unique(M0, return_counts=True)
    big = np.prod(shape0) * 1.0
    for i in np.nonzero(counts > big)[0]:
        M0[M0 == i] = 0

    _, M0 = np.unique(M0, return_inverse=True)
    M0 = np.reshape(M0, shape0)

    return M0


def post_proc_cellpose(
    inst_map: np.ndarray,
    flow_map: np.ndarray,
    dist_map: np.ndarray = None,
    return_flows: bool = False,
    min_size: int = 30,
    **kwargs
) -> np.ndarray:
    """Run the cellpose post-processing pipeline.

    https://www.nature.com/articles/s41592-020-01018-x

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled or binary mask. Shape (H, W).
        flow_map : np.ndarray
            Y- and x-flows. Shape: (2, H, W)
        dist_map : np.ndarray, default=None
            Regressed distance transform. Shape: (H, W).
        return_flows : bool, default=False
            If True, returns the HSV converted flows. They are just not
            needed for anything relevant.
        min_size : int
            The minimum size for the objects that will not be removed.

    Returns
    -------
        np.ndarray:
            The instance labelled segmentation mask. Shape (H, W)
    """
    #  convert channels to CHW
    if dist_map is not None:
        binary_mask = apply_hysteresis_threshold(dist_map, 0.5, 0.5)
    else:
        binary_mask = binarize(inst_map).astype(bool)

    dP = flow_map * binary_mask  # /5.
    # dP = normalize_field(dP)

    pixel_loc, _ = follow_flows(dP, niter=300, mask=binary_mask, suppress_euler=False)

    mask = get_masks_cellpose(p=pixel_loc, mask=binary_mask)
    inst_map = fill_holes_and_remove_small_masks(mask, min_size=min_size).astype("i4")

    if return_flows:
        hsv_flows = gen_flows(dP)
        return inst_map, hsv_flows

    return inst_map



from typing import Tuple

import torch
import torch.nn as nn

from ....modules import Identity
from .cat import CatBlock
from .sum import SumBlock

__all__ = ["Merge"]


MERGE_LOOKUP = {
    "cat": CatBlock,
    "sum": SumBlock,
}


class Merge(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Merge wrapper class.

        Parameters
        ----------
            name : str
                The name of the merging method. One of "sum", "cat".

        Raises
        ------
            ValueError: if the merging method name is illegal.
        """
        super().__init__()

        allowed = list(MERGE_LOOKUP.keys()) + [None]
        if name not in allowed:
            raise ValueError(
                f"Illegal merging method given. Allowed: {allowed}. Got: '{name}'"
            )

        if name is not None:
            self.merge = MERGE_LOOKUP[name](**kwargs)
        else:
            self.merge = Identity(**kwargs)

    @property
    def out_channels(self) -> int:
        """Out channels."""
        out_channels = self.merge.in_channels
        if not isinstance(self.merge, Identity):
            out_channels = self.merge.out_channels

        return out_channels

    def forward(
        self, x: torch.Tensor, skips: Tuple[torch.Tensor, ...], **kwargs
    ) -> torch.Tensor:
        """Forward pass of the merge module."""
        return self.merge(x, skips)



import torch
import torch.nn as nn

__all__ = ["Swish"]


@torch.jit.script
def swish_jit_fwd(input: torch.Tensor) -> torch.Tensor:
    """Swish forward."""
    return input.mul(torch.sigmoid(input))


@torch.jit.script
def swish_jit_bwd(input: torch.Tensor, grad_output: torch.Tensor) -> torch.Tensor:
    """Swish backward."""
    input_s = torch.sigmoid(input)
    return grad_output * (input_s * (1 + input * (1 - input_s)))


class SwishFunction(torch.autograd.Function):
    """Memory efficient Swish implementation."""

    @staticmethod
    def forward(ctx, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        ctx.save_for_backward(x)
        return swish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
        """Backward pass."""
        x = ctx.saved_tensors[0]
        return swish_jit_bwd(x, grad_output)


def swish(x: torch.Tensor) -> torch.Tensor:
    """Apply element-wise swish function."""
    return SwishFunction.apply(x)


class Swish(nn.Module):
    def __init__(self, inplace: bool = False) -> None:
        """Apply the element-wise swish function.

        Parameters
        ----------
            inplace : bool, default=False
                This is not used, exists only for compatibility.
        """
        super().__init__()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass of swish activation.

        Parameters
        ----------
            input : torch.Tensor
                Input tensor. Can be of any shape (C, *)

        Returns
        -------
            torch.Tensor:
                Activated output tensor. Shape same as input.
        """
        return swish(input)



import torch.nn as nn


class GroupNorm(nn.GroupNorm):
    def __init__(self, num_features: int, num_groups: int = None, **kwargs) -> None:
        """Wrap nn.Groupnorm class.

        To make kwargs compatible with nn.BatchNorm.

        Also, infers the `num_groups` from the num_features to avoid
        errors. By default: uses 16 channels per group. If channels <= 16,
        squashes to layer norm.

        Magic number 16 comes from the paper:
        https://arxiv.org/abs/1803.08494

        Parameters
        ----------
            num_features : int
                Number of input channels/features.
            num_groups : int, default=None
                Number of groups to group the channels.
        """
        num_groups, remainder = divmod(num_features, 16)
        if remainder:
            num_groups = num_features // remainder

        super().__init__(num_groups=num_groups, num_channels=num_features)



from typing import Dict, Optional, Tuple

import torch
import torch.nn as nn

from ...decoders import Decoder
from ...modules.misc_modules import StyleReshape
from ..base._base_model import BaseMultiTaskSegModel
from ..base._seg_head import SegHead
from ..base._timm_encoder import TimmEncoder
from ._conf import _create_stardist_args

__all__ = ["StarDistUnet", "stardist_base", "stardist_base_multiclass", "stardist_plus"]


class StarDistUnet(BaseMultiTaskSegModel):
    def __init__(
        self,
        decoders: Tuple[str, ...],
        extra_convs: Dict[str, Dict[str, int]],
        heads: Dict[str, Dict[str, int]],
        inst_key: str = "dist",
        depth: int = 4,
        out_channels: Tuple[int, ...] = (256, 128, 64, 32),
        style_channels: int = None,
        enc_name: str = "resnet50",
        enc_pretrain: bool = True,
        enc_freeze: bool = False,
        long_skip: str = "unet",
        merge_policy: str = "cat",
        short_skip: str = "basic",
        block_type: str = "basic",
        normalization: str = None,
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        attention: str = None,
        preattend: bool = False,
        skip_params: Optional[Dict] = None,
        **kwargs,
    ) -> None:
        """Stardist (2D) U-Net model implementation.

        Stardist:
            - https://arxiv.org/abs/1806.03535

                      (|------ SEMANTIC_DECODER --- CONV --- SEMANTIC_HEAD)
        ENCODER -------|
                       |                                    |---- RAY_HEAD
                       |------ STARDIST_DECODER --- CONV ---|
                                                 |          |---- DISTPROB_HEAD
                                                 |
                                                (|- CONV --- TYPE_HEAD)

        NOTE: Minor differences from the original implementation.
        - long skip concatenation/sum applied before each conv layer rather than after.

        Parameters
        ----------
            decoders : Tuple[str, ...]
                Names of the decoder branches of this network. E.g. ("stardist", "sem")
            extra_convs : Dict[str, Dict[str, int]]
                The extra conv blocks before segmentation heads of the architecture.
                I.e. Names of the decoder branches (has to match `decoders`) mapped to
                dicts of output name - number of output channels. E.g.
                {"stardist": {"type": 128, "stardist": 128}, "sem": {"sem": 128}}
            heads : Dict[str, Dict[str, int]]
                The segmentation heads of the architecture. I.e. Names of the decoder
                branches (has to match `decoders`) mapped to dicts
                 of output name - number of output classes. E.g.
                {"cellpose": {"type": 4, "cellpose": 2}, "sem": {"sem": 5}}
            inst_key : str, default="dist"
                The key for the model output that will be used in the instance
                segmentation post-processing pipeline as the binary segmentation result.
            depth : int, default=4
                The depth of the encoder. I.e. Number of returned feature maps from
                the encoder. Maximum depth = 5.
            out_channels : Tuple[int, ...], default=(256, 128, 64, 32)
                Out channels for each decoder stage.
            style_channels : int, default=256
                Number of style vector channels. If None, style vectors are ignored.
            enc_name : str, default="resnet50"
                Name of the encoder. See timm docs for more info.
            enc_pretrain : bool, default=True
                Whether to use imagenet pretrained weights in the encoder.
            enc_freeze : bool, default=False
                Freeze encoder weights for training.
            long_skip : str, default="unet"
                long skip method to be used. One of: "unet", "unetpp", "unet3p",
                "unet3p-lite", None
            merge_policy : str, default="sum"
                The long skip merge policy. One of: "sum", "cat"
            short_skip : str, default="basic"
                The name of the short skip method. One of: "residual", "dense", "basic"
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=True
                If True, normalization will be applied before convolution.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
            skip_params : Optional[Dict]
                Extra keyword arguments for the skip-connection module. These depend
                on the skip module. Refer to specific skip modules for more info.

        Raises
        ------
            ValueError: If `decoders` does not contain 'stardist'.
            ValueError: If `extra_convs` keys don't match `decoders`.
            ValueError: If extra_convs names don't have a matching head name in `heads`.
        """
        super().__init__()
        self.aux_key = self._check_decoder_args(decoders, ("stardist",))
        self.inst_key = inst_key
        self._check_head_args(extra_convs, decoders)
        self._check_head_args(heads, self._get_inner_keys(extra_convs))
        self._check_depth(depth, {"out_channels": out_channels})

        self.enc_freeze = enc_freeze
        use_style = style_channels is not None
        self.extra_convs = extra_convs
        self.heads = heads

        n_layers = (1,) * depth
        n_blocks = ((2,),) * depth
        dec_params = {
            d: _create_stardist_args(
                depth,
                normalization,
                activation,
                convolution,
                attention,
                preactivate,
                preattend,
                short_skip,
                use_style,
                block_type,
                merge_policy,
                skip_params,
            )
            for d in decoders
        }

        # set timm encoder
        self.encoder = TimmEncoder(
            enc_name,
            depth=depth,
            pretrained=enc_pretrain,
        )

        # style
        self.make_style = None
        if use_style:
            self.make_style = StyleReshape(self.encoder.out_channels[0], style_channels)

        # set decoder
        for decoder_name in decoders:
            decoder = Decoder(
                enc_channels=list(self.encoder.out_channels),
                out_channels=out_channels,
                style_channels=style_channels,
                long_skip=long_skip,
                n_layers=n_layers,
                n_blocks=n_blocks,
                stage_params=dec_params[decoder_name],
            )
            self.add_module(f"{decoder_name}_decoder", decoder)

        # set additional conv blocks ('avoid “fight over features”'.)
        for decoder_name in extra_convs.keys():
            for extra_conv, n_channels in extra_convs[decoder_name].items():
                features = nn.Conv2d(
                    in_channels=decoder.out_channels,
                    out_channels=n_channels,
                    kernel_size=3,
                    padding=1,
                    bias=False,
                )
                self.add_module(f"{extra_conv}_features", features)

        # set heads
        for decoder_name in extra_convs.keys():
            for extra_conv, in_channels in extra_convs[decoder_name].items():
                for output_name, n_classes in heads[extra_conv].items():
                    seg_head = SegHead(
                        in_channels=in_channels,
                        out_channels=n_classes,
                        kernel_size=1,
                    )
                    self.add_module(f"{output_name}_seg_head", seg_head)

        self.name = f"StardistUnet-{enc_name}"

        # init decoder weights
        self.initialize()

        # freeze encoder if specified
        if enc_freeze:
            self.freeze_encoder()

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Forward pass of Stardist."""
        self._check_input_shape(x)

        feats = self.encoder(x)

        style = None
        if self.make_style is not None:
            style = self.make_style(feats[0])

        dec_feats = self.forward_dec_features(feats, style)
        # Extra convs after decoders
        for e in self.extra_convs.keys():
            for extra_conv in self.extra_convs[e].keys():
                k = self.aux_key if extra_conv not in dec_feats.keys() else extra_conv
                dec_feats[extra_conv] = self[f"{extra_conv}_features"](dec_feats[k])

        # seg heads
        for decoder_name in self.heads.keys():
            for head_name in self.heads[decoder_name].keys():
                k = self.aux_key if head_name not in dec_feats.keys() else head_name
                if k != head_name:
                    dec_feats[head_name] = dec_feats[k]

        out = self.forward_heads(dec_feats)

        return out


def stardist_base(n_rays: int, **kwargs) -> nn.Module:
    """Create the Stardist U-net from kwargs.

    Stardist:
    - https://arxiv.org/abs/1806.03535

    Parameters
    ----------
        n_rays : int
            Number of rays predicted per each object
        **kwargs:
            Arbitrary key word args for the StarDistUnet class.

    Returns
    -------
        nn.Module: The initialized Stardist U-net model.
    """
    stardist_unet = StarDistUnet(
        decoders=("stardist",),
        extra_convs={"stardist": {"stardist": 128}},
        heads={"stardist": {"stardist": n_rays, "dist": 1}},
        **kwargs,
    )

    return stardist_unet


def stardist_base_multiclass(n_rays: int, type_classes: int, **kwargs) -> nn.Module:
    """Create the Stardist U-net with an extra multi-class segmentation head.

    Stardist:
    - https://arxiv.org/abs/1806.03535

    Parameters
    ----------
        n_rays : int
            Number of rays predicted per each object
        type_classes : int
            Number of type classes in the dataset.
        **kwargs:
            Arbitrary key word args for the StarDistUnet class.

    Returns
    -------
        nn.Module: The initialized multiclass Stardist U-net model.
    """
    stardist_unet = StarDistUnet(
        decoders=("stardist",),
        extra_convs={"stardist": {"type": 128, "stardist": 128}},
        heads={
            "stardist": {"stardist": n_rays, "dist": 1},
            "type": {"type": type_classes},
        },
        **kwargs,
    )

    return stardist_unet


def stardist_plus(
    n_rays: int, type_classes: int, sem_classes: int, **kwargs
) -> nn.Module:
    """Create the Stardist U-net with a semantic decoder-branch.

    Stardist:
    - https://arxiv.org/abs/1806.03535

    Parameters
    ----------
        n_rays : int
            Number of rays predicted per each object
        type_classes : int
            Number of type classes in the dataset.
        sem_classes : int
            Number of semantic-branch classes.
        **kwargs:
            Arbitrary key word args for the StarDistUnet class.

    Returns
    -------
        nn.Module: The initialized Stardist+ U-net model.
    """
    stardist_unet = StarDistUnet(
        decoders=("stardist", "sem"),
        extra_convs={"stardist": {"type": 128, "stardist": 128}, "sem": {"sem": 128}},
        heads={
            "stardist": {"stardist": n_rays, "dist": 1},
            "type": {"type": type_classes},
            "sem": {"sem": sem_classes},
        },
        **kwargs,
    )

    return stardist_unet



import re
from pathlib import Path

try:
    import requests
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use `SimpleDownloader`, the requests package is needed. "
        "Install with `pip install requests`."
    )
from tqdm import tqdm


class SimpleDownloader:
    """A simple downloader to download data given an url.

    NOTE: Includes downloading from google drive.

    ALSO NOTE: Does not have any fancy functionality. Just downloading..
    """

    @staticmethod
    def gdrive_download(
        file_id: str, file_name: str, save_dir: str, chunk_size: int = 32768
    ) -> None:
        """Download a file from google drive.

        Parameters
        ----------
            file_id : str
                Google drive file ID.
            file_name : str
                Name of the file to be saved. Typically .zip file.
            save_dir : str
                Path to the directory where the file will be saved.
            chunk_size : int, default=32768
                Chunk size for loading the file in bytes. Here 32^3.

        Example
        -------
            >>> save_dir = "/path/to/save_dir"
            >>> gdrive_id = "123gdriveID"
            >>> gdrive_download(gdrive_id, file_name="dd.zip", save_dir=save_dir)
        """
        save_dir = Path(save_dir)
        if not save_dir.exists() or not save_dir.is_dir():
            raise ValueError(
                "The given `save_dir` does not exist or is not a folder." f" {save_dir}"
            )

        path = save_dir / file_name

        url = f"https://drive.google.com/uc?id={file_id}"
        SimpleDownloader._download(
            url, path.as_posix(), gdrive=True, chunk_size=chunk_size
        )

    @staticmethod
    def download(url: str, save_dir: str, chunk_size: int = 32768) -> None:
        """Download a file from url.

        Parameters
        ----------
            url : str
                The url of the file to be downloaded.
            save_dir : str
                Path to the directory where the file will be saved.
            chunk_size : int, default=32768
                Chunk size for loading the file in bytes. Here 32^3.

        Example
        -------
            >>> save_dir = "/path/to/save_dir"
            >>> url = "https://url/to/file/"
            >>> SimpleDownloader.download(url, save_dir) # download from a url.
        """
        save_dir = Path(save_dir)
        if not save_dir.exists() or not save_dir.is_dir():
            raise ValueError(
                "The given `save_dir` does not exist or is not a folder." f" {save_dir}"
            )

        path = save_dir / Path(url).name
        SimpleDownloader._download(
            url, path.as_posix(), gdrive=False, chunk_size=chunk_size
        )

    @staticmethod
    def _download(url: str, out_path: str, gdrive: bool, chunk_size: int) -> None:
        response = SimpleDownloader._get_response(url, gdrive=gdrive)
        nbytes = int(response.headers.get("content-length", 0))

        with tqdm(total=nbytes, unit="iB", unit_scale=True) as pbar, open(
            out_path, "wb"
        ) as f:
            for chunk in response.iter_content(chunk_size):
                f.write(chunk)
                pbar.update(len(chunk))

    @staticmethod
    def _get_response(url: str, gdrive: bool = False) -> requests.Response:
        session = requests.session()
        response = session.get(url, stream=True, verify=True)

        if gdrive:
            url = SimpleDownloader._get_url_from_gdrive_confirmation(response.text)
            response = session.get(url, stream=True, verify=True)

        return response

    @staticmethod
    def _get_url_from_gdrive_confirmation(contents: str) -> str:
        """Parse the response text if furher confirmation is needed.

        From:
        ----
        https://github.com/wkentaro/gdown/blob/main/gdown/download.py
        """
        url = ""
        for line in contents.splitlines():
            m = re.search(r'href="(\/uc\?export=download[^"]+)', line)
            if m:
                url = "https://docs.google.com" + m.groups()[0]
                url = url.replace("&amp;", "&")
                break
            m = re.search('id="downloadForm" action="(.+?)"', line)
            if m:
                url = m.groups()[0]
                url = url.replace("&amp;", "&")
                break
            m = re.search('"downloadUrl":"([^"]+)', line)
            if m:
                url = m.groups()[0]
                url = url.replace("\\u003d", "=")
                url = url.replace("\\u0026", "&")
                break
            m = re.search('<p class="uc-error-subcaption">(.*)</p>', line)
            if m:
                error = m.groups()[0]
                raise RuntimeError(error)
        if not url:
            raise RuntimeError(
                "Cannot retrieve the public link of the file. "
                "You may need to change the permission to "
                "'Anyone with the link', or have had many accesses."
            )
        return url



from .functional.cellpose._old_cellpose import (
    get_masks_cellpose_old,
    post_proc_cellpose_old,
)
from .functional.cellpose.cellpose import get_masks_cellpose, post_proc_cellpose
from .functional.cellpose.utils import (
    fill_holes_and_remove_small_masks,
    gen_flows,
    normalize_field,
)
from .functional.dcan import post_proc_dcan
from .functional.dran import post_proc_dran
from .functional.drfns import post_proc_drfns
from .functional.hovernet import post_proc_hovernet
from .functional.omnipose import get_masks_omnipose, post_proc_omnipose
from .functional.stardist import post_proc_stardist, post_proc_stardist_orig

POSTPROC_LOOKUP = {
    "stardist_orig": post_proc_stardist_orig,
    "stardist": post_proc_stardist,
    "cellpose": post_proc_cellpose,
    "cellpose_old": post_proc_cellpose_old,
    "omnipose": post_proc_omnipose,
    "dcan": post_proc_dcan,
    "drfns": post_proc_drfns,
    "dran": post_proc_dran,
    "hovernet": post_proc_hovernet,
}

__all__ = [
    "POSTPROC_LOOKUP",
    "gen_flows",
    "fill_holes_and_remove_small_masks",
    "normalize_field",
    "get_masks_cellpose",
    "post_proc_cellpose",
    "get_masks_omnipose",
    "post_proc_omnipose",
    "post_proc_hovernet",
    "post_proc_stardist",
    "post_proc_stardist_orig",
    "get_masks_cellpose_old",
    "post_proc_cellpose_old",
    "post_proc_drfns",
    "post_proc_dcan",
    "post_proc_dran",
]



import torch
import torch.nn as nn
import torch.nn.functional as F


# Adapted from:
# https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/norm.py
class LayerNorm2d(nn.LayerNorm):
    def __init__(self, num_features: int, **kwargs) -> None:
        """Layernorm wrap for BCHW shaped tensors.

        Parameters
        ----------
            num_features : int
                Number of input channels/features.
        """
        super().__init__(num_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of 2D layer norm."""
        return F.layer_norm(
            x.permute(0, 2, 3, 1),
            self.normalized_shape,
            self.weight,
            self.bias,
            self.eps,
        ).permute(0, 3, 1, 2)



from typing import Tuple

import numpy as np

__all__ = ["get_stats", "iou_score", "dice_coef", "accuracy", "f1score"]


def get_stats(
    true: np.ndarray, pred: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute basic metrics for diagnostic tests.

    Parameters
    ----------
        true : np.ndarray
            Ground truth binary mask. Shape (H, W)
        pred : np.ndarray
            Predicted binary mask. Shape (H, W)

    Returns
    -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]:
            The computed TP, FP, and FN pixels in a labelled mask.
            Shapes: (H, W)
    """
    tp = true * pred
    fp = np.bitwise_xor(pred, tp)
    fn = np.bitwise_xor(true, tp)

    return tp, fp, fn


def iou_score(
    tp: np.ndarray, fp: np.ndarray, fn: np.ndarray, eps: float = 1e-8
) -> float:
    """Compute the intersection over union (Jaccard-index).

    Parameters
    ----------
        tp : np.ndarray
            True positive pixels. Shape (H, W).
        fp : np.ndarray
            False postive pixels. Shape (H, W).
        fn : np.ndarray
            False negative pixels. Shape (H, W).
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        float:
            The computed intersection over union.
    """
    numerator = tp.sum()
    denominator = fp.sum() + fn.sum() + numerator + eps

    return numerator / denominator


def dice_coef(
    tp: np.ndarray, fp: np.ndarray, fn: np.ndarray, eps: float = 1e-8
) -> float:
    """Compute the Sørensen-Dice coefficient.

    Parameters
    ----------
        tp : np.ndarray
            True positive pixels. Shape (H, W).
        fp : np.ndarray
            False postive pixels. Shape (H, W).
        fn : np.ndarray
            False negative pixels. Shape (H, W).
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        float:
            The computed dice coefficient.
    """
    numerator = 2 * tp.sum()
    denominator = fp.sum() + fn.sum() + numerator + eps

    return numerator / denominator


def accuracy(
    tp: np.ndarray, fp: np.ndarray, fn: np.ndarray, eps: float = 1e-8
) -> float:
    """Compute the binary accuracy.

    Parameters
    ----------
        tp : np.ndarray
            True positive pixels. Shape (H, W).
        fp : np.ndarray
            False postive pixels. Shape (H, W).
        fn : np.ndarray
            False negative pixels. Shape (H, W).
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        float:
            The computed binary accuracy.
    """
    tn = 1 - (tp + fn + fp)
    numerator = tn.sum() + tp.sum()
    denominator = fp.sum() + fn.sum() + numerator + eps

    return numerator / denominator


def f1score(tp: np.ndarray, fp: np.ndarray, fn: np.ndarray, eps: float = 1e-8) -> float:
    """Compute the binary f1-score.

    Parameters
    ----------
        tp : np.ndarray
            True positive pixels. Shape (H, W).
        fp : np.ndarray
            False postive pixels. Shape (H, W).
        fn : np.ndarray
            False negative pixels. Shape (H, W).
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.


    Returns
    -------
        float:
            The computed f1-score.
    """
    numerator = tp.sum()
    denominator = 0.5 * fp.sum() + 0.5 * fn.sum() + numerator + eps

    return numerator / denominator



from cellseg_models_pytorch.inference import ResizeInferer
from cellseg_models_pytorch.models import cellpose_plus
from cellseg_models_pytorch.utils.latency_benchmark import LatencyBenchmarker


def test_latency_benchmark(img_dir):
    model = cellpose_plus(sem_classes=3, type_classes=3, long_skip="unet")

    inferer = ResizeInferer(
        model,
        img_dir,
        out_activations={"sem": "softmax", "type": "softmax", "cellpose": "tanh"},
        out_boundary_weights={"sem": False, "type": False, "cellpose": True},
        resize=(256, 256),
        padding=80,
        instance_postproc="hovernet",
        batch_size=1,
        save_intermediate=True,
        device="cpu",
        parallel=False,
    )
    inferer.infer()

    bm = LatencyBenchmarker(inferer)

    bm.postproc_latency("inst", reps_per_img=1)
    bm.postproc_latency("type", reps_per_img=1)
    bm.postproc_latency("sem", reps_per_img=1)
    bm.inference_latency(reps=1, warmup_reps=0)
    bm.inference_postproc_latency(reps=1)
    # bm.model_latency(input_size=(64, 64), reps=1, warmup_reps=0, device="cpu")
    # bm.model_throughput(input_size=(64, 64), reps=1, warmup_reps=0, device="cpu")



from pathlib import Path
from typing import Union

import cv2
import numpy as np
import scipy.io as sio

from .mask_utils import bounding_box, fix_duplicates, get_inst_centroid, get_inst_types


class FileHandler:
    """Class for handling file I/O."""

    @staticmethod
    def read_img(path: Union[str, Path]) -> np.ndarray:
        """Read an image & convert from bgr to rgb. (cv2 reads imgs in bgr).

        Parameters
        ----------
            path : str or Path
                Path to the image file.

        Returns
        -------
            np.ndarray:
                The image. Shape (H, W, 3).
        """
        path = Path(path)
        return cv2.cvtColor(cv2.imread(path.as_posix()), cv2.COLOR_BGR2RGB)

    @staticmethod
    def write_img(path: Union[str, Path], img: np.ndarray) -> None:
        """Write an image.

        Parameters
        ----------
            path : str or Path
                Path to the image file.
            img : np.ndarray
                The image to be written.

        """
        path = Path(path)
        cv2.imwrite(path.as_posix(), cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

    @staticmethod
    def read_mask(
        path: Union[str, Path],
        key: str = "inst_map",
        retype: bool = True,
        return_all: bool = False,
    ) -> Union[np.ndarray, None]:
        """Read a mask from a .mat file.

        If a mask is not found, return None

        Parameters
        ----------
            path : str or Path
                Path to the image file.
            key : str, default="inst_map"
                Name/key of the mask type that is being read from .mat
            retype : bool, default=True
                Convert the matrix type.
            return_all : bool, default=False
                Return the whole dict. Overrides the `key` arg.

        Returns
        -------
            np.ndarray or None:
                The mask indice matrix. Shape (H, W)

        Raises
        ------
            ValueError: If an illegal key is given.
        """
        dtypes = {
            "inst_map": "int32",
            "type_map": "int32",
            "sem_map": "int32",
            "inst_centroid": "float64",
            "inst_type": "int32",
        }

        path = Path(path)
        if not path.exists():
            raise ValueError(f"{path} not found")

        try:
            mask = sio.loadmat(path.as_posix())
        except Exception:
            mask = None

        if not return_all:
            allowed = ("inst_map", "type_map", "inst_centroid", "inst_type", "sem_map")
            if key not in allowed:
                raise ValueError(f"Illegal key given. Got {key}. Allowed: {allowed}")

            try:
                mask = mask[key]
                if retype:
                    mask = mask.astype(dtypes[key])
            except Exception:
                mask = None

        return mask

    @staticmethod
    def write_mask(
        path: Union[str, Path],
        inst_map: np.ndarray,
        type_map: np.ndarray = None,
        sem_map: np.ndarray = None,
        compute_centorids: bool = False,
        compute_bboxes: bool = False,
    ) -> None:
        """
        Write multiple masks to .mat file.

        Parameters
        ----------
            path : str or Path
                Path to the .mat file.
            inst_map : np.ndarray
                The inst_map to be written.
            type_map : np.ndarray, optional
                The inst_map to be written.
            sem_map : np.ndarray, optional
                The inst_map to be written.
            compute_centroids : bool, optional
                Flag to tompute instance centorids.
            compute_bboxes : bool, optional
                Flag to tompute instance bboxes.
        """
        path = Path(path)
        if not path.suffix == ".mat":
            raise ValueError(f"File suffix needs to be '.mat'. Got {path.suffix}.")

        inst_map = fix_duplicates(inst_map)
        inst_types = get_inst_types(inst_map, type_map)

        res = {
            "inst_map": inst_map,
            "inst_type": inst_types,
        }

        if compute_centorids:
            centroids = get_inst_centroid(inst_map)
            res["inst_centroid"] = centroids

        if compute_bboxes:
            inst_ids = list(np.unique(inst_map)[1:])
            bboxes = np.array(
                [bounding_box(np.array(inst_map == id_, np.uint8)) for id_ in inst_ids]
            )
            res["inst_bbox"] = bboxes

        if type_map is not None:
            res["type_map"] = type_map

        if sem_map is not None:
            res["sem_map"] = type_map

        sio.savemat(
            file_name=path,
            mdict=res,
        )



import numpy as np
import pytest

from cellseg_models_pytorch.transforms.albu_transforms import (
    compose,
    imgnorm_transform,
    minmaxnorm_transform,
    percentilenorm_transform,
)


@pytest.mark.parametrize(
    "transform",
    [imgnorm_transform, percentilenorm_transform, minmaxnorm_transform],
)
@pytest.mark.parametrize("zero_input", [None, np.zeros((14, 14, 3), dtype=np.float32)])
def test_norm_transforms(img_sample, transform, zero_input):
    trans = compose([transform()])

    if zero_input is not None:
        img_sample = zero_input

    transformed = trans(image=img_sample)

    assert transformed["image"].dtype == np.float32
    assert transformed["image"].shape == img_sample.shape



from typing import List

import numpy as np
from albumentations.core.transforms_interface import ImageOnlyTransform

from ...utils import minmax_normalize, normalize, percentile_normalize

__all__ = ["imgnorm_transform", "percentilenorm_transform", "minmaxnorm_transform"]


class MinMaxNormalization(ImageOnlyTransform):
    def __init__(
        self,
        amin: float = None,
        amax: float = None,
        always_apply: bool = True,
        p: float = 1.0,
        **kwargs
    ) -> None:
        """Min-max normalization transformation.

        Parameters
        ----------
            amin : float, optional
                Clamp min value. No clamping performed if None.
            amax : float, optional
                Clamp max value. No clamping performed if None.
        """
        super().__init__(always_apply, p)
        self.amin = amin
        self.amax = amax

    def apply(self, image: np.ndarray, **kwargs) -> np.ndarray:
        """Apply min-max normalization.

        Parameters
        ----------
            image : np.ndarray:
                Input image to be normalized. Shape (H, W, C)|(H, W).

        Returns
        -------
            np.ndarray:
                Normalized image. Same shape as input. dtype: float32.
        """
        return minmax_normalize(image, self.amin, self.amax)


class PercentileNormalization(ImageOnlyTransform):
    def __init__(
        self,
        lower: float = 0.01,
        upper: float = 99.99,
        always_apply: bool = True,
        p: float = 1.0,
        **kwargs
    ) -> None:
        """Percentile normalization transformation.

        Parameters
        ----------
            amin : float, optional
                Clamp min value. No clamping performed if None.
            amax : float, optional
                Clamp max value. No clamping performed if None.
        """
        super().__init__(always_apply, p)
        self.lower = lower
        self.upper = upper

    def apply(self, image: np.ndarray, **kwargs) -> np.ndarray:
        """Apply percentile normalization to input image.

        Parameters
        ----------
            image : np.ndarray:
                Input image to be normalized. Shape (H, W, C)|(H, W).

        Returns
        -------
            np.ndarray:
                Normalized image. Same shape as input. dtype: float32.
        """
        return percentile_normalize(image, self.lower, self.upper)


class ImgNormalization(ImageOnlyTransform):
    def __init__(
        self,
        standardize: bool = True,
        amin: float = None,
        amax: float = None,
        always_apply: bool = True,
        p: float = 1.0,
        **kwargs
    ) -> None:
        """Image level normalization transformation.

        NOTE: this is not dataset-level normalization but image-level.

        Parameters
        ----------
            standardize : bool, default=True
                If True, divides the mean shifted img by the standard deviation.
            amin : float, optional
                Clamp min value. No clamping performed if None.
            amax : float, optional
                Clamp max value. No clamping performed if None.
        """
        super().__init__(always_apply, p)
        self.standardize = standardize
        self.amin = amin
        self.amax = amax

    def apply(self, image: np.ndarray, **kwargs) -> np.ndarray:
        """Apply image-level normalization to input image.

        Parameters
        ----------
            image : np.ndarray:
                Input image to be normalized. Shape (H, W, C)|(H, W).

        Returns
        -------
            np.ndarray:
                Normalized image. Same shape as input. dtype: float32.
        """
        return normalize(image, self.standardize, self.amin, self.amax)


def imgnorm_transform(**kwargs) -> List[ImageOnlyTransform]:
    """Return image-level normalization transform."""
    return [ImgNormalization(**kwargs)]


def percentilenorm_transform(**kwargs) -> List[ImageOnlyTransform]:
    """Return percentile normalization transform."""
    return [PercentileNormalization(**kwargs)]


def minmaxnorm_transform(**kwargs) -> List[ImageOnlyTransform]:
    """Return minmax normalization transform."""
    return [MinMaxNormalization(**kwargs)]



from typing import Callable, Tuple

import pytest
import scipy.ndimage as ndi
import torch

from cellseg_models_pytorch.losses import Loss
from cellseg_models_pytorch.transforms.functional import gen_weight_maps


def _get_dummy_pair(n_classes: int) -> Tuple[torch.Tensor]:
    dummy_yhat = torch.zeros([1, n_classes, 64, 64]).float()
    dummy_target = torch.zeros([1, 64, 64]).long()

    for i, c in enumerate(range(dummy_yhat.shape[1]), 1):
        dummy_yhat[
            ..., c, (i * c) + 10 : (i * c) + 20, (i * c) + 30 : (i * c) + 40
        ] += 1
        dummy_yhat[
            ..., c, (i * c) + 10 : (i * c) + 20, (i * c) + 30 : (i * c) + 40
        ] += 1
        dummy_target[
            ...,
            (i * c * 2) + 10 : (i * c * 2) + 20,
            (i * c * 2) + 30 : (i * c * 2) + 40,
        ] += 1
        dummy_target[
            ...,
            (i * c * 2) + 15 : (i * c * 2) + 25,
            (i * c * 2) + 15 : (i * c * 2) + 25,
        ] += 1
        dummy_target[dummy_target > c] = c

    return dummy_yhat, dummy_target


def _get_dummy_weight_map(target: torch.Tensor) -> torch.Tensor:
    wmap = gen_weight_maps(ndi.label(target)[0][0])
    wmap = torch.tensor(wmap).unsqueeze(0)
    return wmap


def _get_dummy_class_weights(n_classes: int) -> torch.Tensor:
    w = torch.zeros(1, n_classes)
    w[0][0] += 0.1

    for i in range(1, n_classes):
        w[0][i] += 0.9 + i * 0.01

    return w[0, ...]


@torch.no_grad()
@pytest.mark.parametrize("n_classes", [1, 3])
@pytest.mark.parametrize("class_weights", [True, False])
@pytest.mark.parametrize("edge_weight", [1.1, None])
@pytest.mark.parametrize("apply_ls", [True, False])
@pytest.mark.parametrize("apply_svls", [True, False])
@pytest.mark.parametrize("apply_sd", [True, False])
@pytest.mark.parametrize(
    "loss",
    ["mse", "ssim", "msssim", "dice", "iou", "focal", "tversky", "ce", "sce"],
)
def test_loss(
    n_classes: int,
    loss: Callable,
    edge_weight: float,
    class_weights: bool,
    apply_svls: bool,
    apply_ls: bool,
    apply_sd: bool,
) -> None:
    yhat, target = _get_dummy_pair(n_classes)

    wmap = None
    if edge_weight is not None:
        wmap = _get_dummy_weight_map(target)

    cw = None
    if class_weights:
        cw = _get_dummy_class_weights(n_classes)

    criterion = Loss(
        name=loss,
        apply_sd=apply_sd,
        apply_ls=apply_ls,
        apply_svls=apply_svls,
        edge_weight=edge_weight,
        class_weights=cw,
    )
    criterion(yhat=yhat, target=target, target_weight=wmap)



from typing import Dict, Optional, Tuple, Union

import torch
import torch.nn as nn

from ..modules.base_modules import Up
from ..modules.conv_layer import ConvLayer
from .long_skips import LongSkip

__all__ = ["DecoderStage"]


class DecoderStage(nn.Module):
    def __init__(
        self,
        stage_ix: int,
        dec_channels: Tuple[int, ...],
        dec_dims: Tuple[int, ...],
        skip_channels: Tuple[int, ...],
        style_channels: int = None,
        n_layers: int = 1,
        n_blocks: Tuple[int, ...] = (2,),
        short_skips: Tuple[str, ...] = ("residual",),
        expand_ratios: Tuple[float, float] = ((1.0, 1.0),),
        block_types: Tuple[Tuple[str, ...], ...] = (("basic", "basic"),),
        normalizations: Tuple[Tuple[str, ...], ...] = (("bn", "bn"),),
        activations: Tuple[Tuple[str, ...], ...] = (("relu", "relu"),),
        convolutions: Tuple[Tuple[str, ...], ...] = (("conv", "conv"),),
        attentions: Tuple[Tuple[str, ...], ...] = ((None, "se"),),
        preactivates: Tuple[Tuple[bool, ...], ...] = ((False, False),),
        preattends: Tuple[Tuple[bool, ...], ...] = ((False, False),),
        use_styles: Tuple[Tuple[bool, ...], ...] = ((False, False),),
        kernel_sizes: Tuple[Tuple[int, ...]] = ((3, 3),),
        groups: Tuple[Tuple[int, ...]] = ((1, 1),),
        biases: Tuple[Tuple[bool, ...]] = ((False, False),),
        upsampling: str = "fixed-unpool",
        long_skip: str = "unet",
        merge_policy: str = "sum",
        layer_residual: bool = False,
        skip_params: Optional[Dict] = None,
        **kwargs,
    ) -> None:
        """Build a decoder stage.

        Operations in each decoder stage:
        1. Upsample
        2. Long skip from encoder to decoder if not the last stage.
        3. Conv block

        Parameters
        ----------
            stage_ix : int
                The index number of the current decoder stage.
            dec_channels : Tuple[int, ...]
                The number of output channels in the decoder output stages. First elem
                is the number of channels in the encoder head or bottleneck.
            dec_dims : Tuple[int, ...]
                Tuple of the heights/widths of each encoder/decoder feature map
                e.g. (8, 16, 32, 64, 128, 256). Feature maps are assumed to be square.
            skip_channels : Tuple[int, ...]
                List of the number of channels in the encoder skip tensors. Ignored if
                `long_skip` == None.
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
            n_layers : int, default=1
                The number of conv layers inside one decoder stage.
            n_blocks : int, default=2
                Number of conv-blocks inside one conv layer.
            short_skips : str, default=("residual", )
                The short skip methods used inside the conv layers.
            expand_ratios : Tuple[float, ...], default=((1.0, 1.0),):
                Expansion/Squeeze ratios for the out channels of each conv block.
            block_types : Tuple[Tuple[str, ...], ...], default=(("basic", "basic"), )
                The type of the convolution blocks in the conv blocks inside the layers.
            normalizations : Tuple[Tuple[str, ...], ...], default: (("bn", "bn"), )
                Normalization methods used in the conv blocks inside the conv layers.
            activations : Tuple[Tuple[str, ...], ...], default: (("relu", "relu"), )
                Activation methods used inside the conv layers.
            attentions : Tuple[Tuple[str, ...], ...], default: ((None, "se"), )
                Attention methods used inside the conv layers.
            preactivates Tuple[Tuple[bool, ...], ...], default: ((False, False), )
                Boolean flags for the conv layers to use pre-activation.
            preattends Tuple[Tuple[bool, ...], ...], default: ((False, False), )
                Boolean flags for the conv layers to use pre-activation.
            use_styles : Tuple[Tuple[bool, ...], ...], default=((False, False), )
                Boolean flags for the conv layers to add style vectors at each block.
            kernel_sizes : Tuple[int, ...], default=((3, 3),)
                The size of the convolution kernels in each conv block.
            groups : int, default=((1, 1),)
                Number of groups for the kernels in each convolution blocks.
            biases : bool, default=((False, False),)
                Include bias terms in the convolution blocks.
            upsampling : str, default="fixed-unpool"
                Name of the upsampling method.
            long_skip : str, default="unet"
                long skip method to be used. One of: "unet", "unetpp", "unet3p",
                 "unet3p-lite", None,
            merge_policy : str, default="sum"
                The long skip merge policy. One of: "sum", "cat"
            layer_residual : bool, default=False
                Apply a layer level residual skip at each layer. I.e x + layer(x)
            skip_params : Optional[Dict]
                Extra keyword arguments for the skip-connection module. These depend
                on the skip module. Refer to specific skip modules for more info.

        Raises
        ------
            ValueError:
                If lengths of the tuple arguments are not equal to `n_layers`.
        """
        super().__init__()

        illegal_args = [
            (k, a)
            for k, a in locals().items()
            if isinstance(a, tuple)
            and a not in (skip_channels, dec_channels, dec_dims)
            and len(a) != n_layers
        ]

        if illegal_args:
            raise ValueError(
                f"""
                All the tuple-arg lengths need to be equal to `n_layers`={n_layers}.
                Illegal args: {illegal_args}"""
            )

        self.long_skip = long_skip
        self.stage_ix = stage_ix
        self.in_channels = dec_channels[stage_ix]
        self.out_channels = dec_channels[stage_ix + 1]

        # upsampling method
        self.upsample = Up(upsampling)

        # long skip connection method
        self.skip = LongSkip(
            name=long_skip,
            merge_policy=merge_policy,
            stage_ix=self.stage_ix,
            in_channels=self.in_channels,
            dec_channels=dec_channels,
            skip_channels=skip_channels,
            dec_dims=dec_dims,
            **skip_params if skip_params is not None else {"k": None},
        )

        # Set up n layers of conv blocks
        layer = None  # placeholder
        self.conv_layers = nn.ModuleDict()
        for i in range(n_layers):
            n_in_feats = self.skip.out_channels if i == 0 else layer.out_channels
            layer = ConvLayer(
                in_channels=n_in_feats,
                out_channels=self.out_channels,
                n_blocks=n_blocks[i],
                layer_residual=layer_residual,
                style_channels=style_channels,
                short_skip=short_skips[i],
                expand_ratios=expand_ratios[i],
                block_types=block_types[i],
                normalizations=normalizations[i],
                activations=activations[i],
                convolutions=convolutions[i],
                preactivates=preactivates[i],
                preattends=preattends[i],
                use_styles=use_styles[i],
                attentions=attentions[i],
                kernel_sizes=kernel_sizes[i],
                groups=groups[i],
                biases=biases[i],
                **kwargs,
            )
            self.conv_layers[f"conv_layer{i + 1}"] = layer

        self.out_channels = layer.out_channels

    def forward(
        self,
        x: torch.Tensor,
        skips: Tuple[torch.Tensor, ...],
        extra_skips: Tuple[torch.Tensor, ...] = None,
        style: torch.Tensor = None,
    ) -> Tuple[torch.Tensor, Union[None, Tuple[torch.Tensor, ...]]]:
        """Forward pass of the decoder stage.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Shape (B, C, H, W).
            skips : Tuple[torch.Tensor, ...]
                All of feature maps from consecutive encoder blocks.
                Order is bottom up. Shapes: (B, C, H, W).
            extra_skips : Tuple[torch.Tensor, ...], default=None
                Extra skip connections. Used in unet3+ and unet++.
            style : torch.Tensor
                Style vector. Shape (B, C).

        Returns
        -------
            Tuple[torch.Tensor, Union[None, Tuple[torch.Tensor, ...]]]:
                Output torch.Tensor and extra skip torch.Tensors. If no extra
                skips are present, returns None as the second return value.
        """
        x = self.upsample(x)

        # long skip
        x = self.skip(x, ix=self.stage_ix, skips=skips, extra_skips=extra_skips)

        # unetpp returns extra skips
        extra_skips = x[1] if self.long_skip == "unetpp" else None
        x = x[0] if self.long_skip == "unetpp" else x

        # conv layer
        for _, conv_layer in self.conv_layers.items():
            x = conv_layer(x, style)

        return x, extra_skips



import pytest
import torch

from cellseg_models_pytorch.modules import ConvBlock


@pytest.mark.parametrize(
    "convblock", ["basic", "mbconv", "fmbconv", "dws", "bottleneck", "hover_dense"]
)
@pytest.mark.parametrize("short_skip", ["residual", "dense", "basic"])
@pytest.mark.parametrize("preattend", [True, False])
@pytest.mark.parametrize("preactivate", [True, False])
@pytest.mark.parametrize("style_channels", [None, 4])
@pytest.mark.parametrize("out_channels", [4, 8])
def test_conv_block_fwdbwd(
    convblock,
    short_skip,
    preattend,
    preactivate,
    out_channels,
    style_channels,
):
    input = torch.rand([1, 8, 16, 16])

    style = None
    if style_channels is not None:
        style = torch.ones([1, style_channels])

    block = ConvBlock(
        name=convblock,
        short_skip=short_skip,
        in_channels=8,
        out_channels=out_channels,
        style_channels=style_channels,
        attention="gc",
        preactivate=preactivate,
        preattend=preattend,
        expand_ratio=1.0,
    )

    output = block(input, style)
    output.mean().backward()

    assert output.shape == torch.Size([1, out_channels, 16, 16])
    assert output.dtype == input.dtype



import torch
import torch.nn.functional as F

from cellseg_models_pytorch.utils import tensor_one_hot

from ..weighted_base_loss import WeightedBaseLoss


class SCELoss(WeightedBaseLoss):
    def __init__(
        self,
        alpha: float = 0.5,
        beta: float = 1.0,
        apply_sd: bool = False,
        apply_ls: bool = False,
        apply_svls: bool = False,
        edge_weight: float = None,
        class_weights: torch.Tensor = None,
        **kwargs
    ) -> None:
        """Symmetric Cross Entropy loss.

        https://arxiv.org/abs/1908.06112

        Parameters
        ----------
            alpha : float, default=0.5
                Weight factor b/w [0,1].
            beta :float, default=1.0
                Weight factor b/w [0,1].
            apply_sd : bool, default=False
                If True, Spectral decoupling regularization will be applied  to the
                loss matrix.
            apply_ls : bool, default=False
                If True, Label smoothing will be applied to the target.
            apply_svls : bool, default=False
                If True, spatially varying label smoothing will be applied to the target
            edge_weight : float, default=none
                Weight that is added to object borders.
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (n_classes,).
        """
        super().__init__(apply_sd, apply_ls, apply_svls, class_weights, edge_weight)
        self.alpha = alpha
        self.beta = beta
        self.eps = 1e-6

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_weight: torch.Tensor = None,
        **kwargs
    ) -> torch.Tensor:
        """Compute the symmetric cross entropy loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).
            target_weight : torch.Tensor, default=None
                The edge weight map. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed SCE loss (scalar).
        """
        num_classes = yhat.shape[1]
        target_one_hot = tensor_one_hot(target, num_classes)
        yhat_soft = F.softmax(yhat, dim=1) + self.eps
        assert target_one_hot.shape == yhat.shape

        yhat = torch.clamp(yhat_soft, min=1e-7, max=1.0)
        target_one_hot = torch.clamp(target_one_hot, min=1e-4, max=1.0)

        forward = target_one_hot * torch.log(yhat_soft)
        reverse = yhat_soft * torch.log(target_one_hot)

        cross_entropy = -torch.sum(forward, dim=1)  # to (B, H, W)
        reverse_cross_entropy = -torch.sum(reverse, dim=1)  # to (B, H, W)

        if self.class_weights is not None:
            cross_entropy = self.apply_class_weights(cross_entropy, target)
            reverse_cross_entropy = self.apply_class_weights(
                reverse_cross_entropy, target
            )

        if self.edge_weight is not None:
            cross_entropy = self.apply_edge_weights(cross_entropy, target_weight)
            reverse_cross_entropy = self.apply_edge_weights(
                reverse_cross_entropy, target_weight
            )

        loss = (
            self.alpha * cross_entropy.mean() + self.beta * reverse_cross_entropy.mean()
        )

        if self.apply_sd:
            loss = self.apply_spectral_decouple(loss, yhat)

        return loss



import pytest
import torch

from cellseg_models_pytorch.decoders.long_skips.merging import Merge


@pytest.mark.parametrize("name", ["sum", "cat", None])
def test_merge(name):
    t1 = torch.rand([1, 1, 16, 16])
    t2 = torch.rand([1, 2, 16, 16])
    t3 = torch.rand([1, 3, 16, 16])

    merge = Merge(name, in_channels=1, skip_channels=(2, 3))
    out = merge(t1, (t2, t3))

    assert out.shape[1] == merge.out_channels



"""
Spatial attention modules for CNNs.

Most of these are simplified and adapted from:
https://github.com/rwightman/pytorch-image-models

License:

Copyright 2019 Ross Wightman

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.models.layers.helpers import make_divisible

from .base_modules import Activation, Conv, Identity, Norm

__all__ = [
    "Attention",
    "SqueezeAndExcite",
    "SCSqueezeAndExcite",
    "ECA",
    "GlobalContext",
]


class SqueezeAndExcite(nn.Module):
    def __init__(
        self,
        in_channels: int,
        squeeze_ratio: float = 0.25,
        conv: str = "conv",
        activation: str = "relu",
        gate_activation: str = "sigmoid",
        **kwargs,
    ) -> None:
        """Squeeze-and-Excitation block.

        https://arxiv.org/abs/1709.01507

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            squeeze_ratio : float, default=0.25
                Ratio of squeeze.
            conv : str, default="conv"
                Convolution layer type.
            activation : str, default="relu"
                Activation layer after squeeze.
            gate_activation : str, default="sigmoid"
                Attention gate function.
        """
        super().__init__()

        squeeze_channels = round(in_channels * squeeze_ratio)

        if squeeze_channels < 2:
            squeeze_channels = 2

        # squeeze channel pooling
        self.conv_squeeze = Conv(
            conv,
            in_channels=in_channels,
            out_channels=squeeze_channels,
            kernel_size=1,
            bias=True,
            padding=0,
        )
        self.act = Activation(activation)

        # excite channel pooling
        self.conv_excite = Conv(
            conv,
            in_channels=squeeze_channels,
            out_channels=in_channels,
            kernel_size=1,
            bias=True,
            padding=0,
        )
        self.gate = Activation(gate_activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """SE-forward pass."""
        # squeeze
        x_se = x.mean((2, 3), keepdim=True)
        x_se = self.conv_squeeze(x_se)
        x_se = self.act(x_se)

        # excite
        x_se = self.conv_excite(x_se)

        return x * self.gate(x_se)


class SCSqueezeAndExcite(SqueezeAndExcite):
    def __init__(
        self,
        in_channels: int,
        squeeze_ratio: float = 0.25,
        conv: str = "conv",
        activation: str = "relu",
        gate_activation: str = "sigmoid",
        **kwargs,
    ) -> None:
        """Spatial and Channel Squeeze & Excitation.

        https://arxiv.org/abs/1803.02579

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            squeeze_ratio : float, default=0.25
                Ratio of squeeze.
            conv : str, default="conv"
                Convolution layer type.
            activation : str, default="relu"
                Activation layer after squeeze.
            gate_activation : str, default="sigmoid"
                Attention gate function.
        """
        super().__init__(
            in_channels=in_channels,
            squeeze_ratio=squeeze_ratio,
            conv=conv,
            activation=activation,
            gate_activation=gate_activation,
        )

        # self.conv_squeeze2 = nn.Conv2d(in_channels, 1, 1)
        self.conv_squeeze2 = Conv(
            conv, in_channels=in_channels, out_channels=1, kernel_size=1, padding=0
        )
        self.gate2 = Activation(gate_activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """SCSE forward pass."""
        # Channel-attention
        x_ce = x.mean((2, 3), keepdim=True)
        x_ce = self.conv_squeeze(x_ce)
        x_ce = self.act(x_ce)
        x_ce = self.conv_excite(x_ce)

        # Spatial attention
        x_se = self.conv_squeeze2(x)

        return x * self.gate(x_ce) + x * self.gate2(x_se)


class ECA(nn.Module):
    def __init__(
        self,
        in_channels: int,
        beta: int = 1,
        gamma: int = 2,
        gate_activation: str = "sigmoid",
        **kwargs,
    ) -> None:
        """Efficient Channel Attention (ECA).

        https://arxiv.org/abs/1910.03151

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            beta : int, default=1
                Coefficient used to compute the kernel size adaptively.
            gamma : int, default=2
                Coefficient used to compute the kernel size adaptively.
            gate_activation : str, default="sigmoid"
                Attention gate function.
        """
        super().__init__()

        # Compute the adaptive kernel size
        t = int(abs(math.log(in_channels, 2) + beta) / gamma)
        kernel_size = max(t if t % 2 else t + 1, 3)

        padding = (kernel_size - 1) // 2
        self.conv = nn.Conv1d(
            1, 1, kernel_size=kernel_size, padding=padding, bias=False
        )

        self.gate = Activation(gate_activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ECA Forward pass."""
        B = x.shape[0]

        y = x.mean((2, 3)).view(B, 1, -1)
        y = self.conv(y)
        y = self.gate(y).view(B, -1, 1, 1)

        return x * y.expand_as(x)


class GlobalContext(nn.Module):
    def __init__(
        self,
        in_channels: int,
        squeeze_ratio: float = 0.125,
        conv: str = "conv",
        activation: str = "relu",
        **kwargs,
    ) -> None:
        """Global context attention block.

        https://arxiv.org/abs/1904.11492

        NOTE: Only the (attn + add)-fusion (the best) variant implemented.

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            squeeze_ratio : float, default=0.25
                Ratio of squeeze.
            conv : str, default="conv"
                Convolution layer type.
            activation : str, default="relu"
                Activation layer after squeeze.
        """
        super().__init__()

        self.conv_attn = Conv(
            conv,
            in_channels=in_channels,
            out_channels=1,
            kernel_size=1,
            padding=0,
            bias=True,
        )
        squeeze_channels = make_divisible(in_channels * squeeze_ratio)

        self.conv_squeeze = Conv(
            conv,
            in_channels=in_channels,
            out_channels=squeeze_channels,
            kernel_size=1,
            padding=0,
            bias=True,
        )
        self.norm = Norm("ln", num_features=squeeze_channels)
        self.act = Activation(activation)
        self.conv_excite = Conv(
            conv,
            in_channels=squeeze_channels,
            out_channels=in_channels,
            kernel_size=1,
            padding=0,
            bias=True,
        )

        self.reset_parameters()

    def reset_parameters(self):
        """Init conv kernel weights."""
        nn.init.kaiming_normal_(
            self.conv_attn.conv.weight, mode="fan_in", nonlinearity="relu"
        )
        nn.init.zeros_(self.conv_excite.conv.weight)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of global context block."""
        B, C, H, W = x.shape

        # Query independent global context attention
        attn = self.conv_attn(x).reshape(B, 1, H * W)  # (B, 1, H * W)
        attn = F.softmax(attn, dim=-1).unsqueeze(3)  # (B, 1, H * W, 1)
        context = x.reshape(B, C, H * W).unsqueeze(1) @ attn
        context = context.view(B, C, 1, 1)

        # squeeze and excite for the global context.
        context = self.conv_squeeze(context)
        context = self.norm(context)
        context = self.act(context)
        context = self.conv_excite(context)

        # fuse (add)
        x = x + context

        return x


ATT_LOOKUP = {
    "se": SqueezeAndExcite,
    "scse": SCSqueezeAndExcite,
    "eca": ECA,
    "gc": GlobalContext,
}


class Attention(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Attention wrapper class.

        Parameters:
        -----------
            name : str
                Name of the attention method.
        """
        super().__init__()

        allowed = list(ATT_LOOKUP.keys()) + [None]
        if name not in allowed:
            raise ValueError(
                f"Illegal attention method given. Allowed: {allowed}. Got: '{name}'"
            )

        if name is not None:
            self.att = ATT_LOOKUP[name](**kwargs)
        else:
            self.att = Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the attention method."""
        return self.att(x)



from . import inference, models, utils
from .models import CellPoseUnet, HoverNet, StarDistUnet

__version__ = "0.1.4"
submodules = ["utils", "models", "inference"]
__all__ = [
    "__version__",
    "CellPoseUnet",
    "StarDistUnet",
    "HoverNet",
    "utils",
    "models",
    "inference",
]



"""
Imported from the cellpose repo.

https://github.com/MouseLand/cellpose/blob/master/cellpose/dynamics.py

Removed unnecessary stuff + code
style mods + docstrings for readability.

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from typing import Tuple

import numpy as np
import torch

__all__ = ["steps2D_interp", "follow_flows"]


def steps2D_interp(
    p: np.ndarray,
    dP: np.ndarray,
    niter: int = 200,
    suppress_euler: bool = False,
) -> np.ndarray:
    """Run Euler integration.

    "run a dynamical system starting at that pixel location and
     following the spatial derivatives specified by the horizontal and
     vertical gradient maps. We use finite differences with a step size
     of one."

    Parameters
    ----------
        p : np.ndarray
            All the pixel locations for the pixels in the flow maps.
            Including zero-pixels. Shape: (2, H, W). Dtype: float32
        dP : np.ndarray
            Flow maps. Shape: (2, H, W). Dtype: float64
        niter : int, default=200
            Number of iterations of dynamics to run

    Returns
    -------
        np.ndarray:
            The pixel locations after dynamics. Shape (2, H, W).

    """
    device = "cuda" if torch.cuda.is_available() else "cpu"

    shape = dP.shape[1:]
    shape = np.array(shape)[[1, 0]].astype("double") - 1

    # Flip dims and modify to shape [1, 1, H*W, 2] for `grid_sample`
    pt = torch.from_numpy(p[[1, 0]].T).double().to(device)
    pt = pt.unsqueeze(0).unsqueeze(0)

    im = torch.from_numpy(dP[[1, 0]]).double().to(device)
    im = im.unsqueeze(0)

    # normalize pt between 0 and 1, normalize the flow
    for k in range(2):
        im[:, k, ...] *= 2.0 / shape[k]
        pt[..., k] /= shape[k]

    # normalize to between -1 and 1
    pt = pt * 2 - 1

    # Euler integration steps
    for t in range(niter):
        dPt = torch.nn.functional.grid_sample(im, pt, align_corners=False)

        if suppress_euler:
            dPt /= 1 + t

        # clamp the final pixel locations
        for k in range(2):
            pt[..., k] = torch.clamp(pt[..., k] + dPt[:, k, ...], -1.0, 1.0)

    # Undo the normalization from before, reverse order of operations
    pt = (pt + 1) * 0.5
    for k in range(2):
        pt[..., k] *= shape[k]

    return pt[..., [1, 0]].cpu().numpy().squeeze().T


def follow_flows(
    dP: np.ndarray,
    mask: np.ndarray = None,
    niter: int = 200,
    suppress_euler: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """Define pixels and run dynamics to recover masks in 2D.

    Pixels are meshgrid. Only pixels with non-zero cell-probability
    are used (as defined by inds).

    Parameters
    ----------
        dP : np.ndarray
            Flow maps. Shape: (2, H, W). Dtype: float64.
        mask : np.ndarray, default=None
            Pixel mask to seed masks. Useful when flows have low magnitudes.
        niter : int, default=200
            Number of iterations of dynamics to run.
        suppress_euler : bool, default=False
            Suppression factor for the euler intergator. Used for
            omnipose.

    Returns
    ---------------
        Tuple[np.ndarray, np.ndarray]:
            A tuple of nd.arrays. The first index is teh final locations
            of each pixel after dynamics (Shape: (2, H, W)) The second
            index is the indices of the non-zero pixels.
            Shape: (number of non zero pixels, 2)
    """
    shape = np.array(dP.shape[1:]).astype(np.int32)
    niter = np.uint32(niter)

    p = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing="ij")
    p = np.array(p).astype(np.float32)

    # run dynamics on subset of pixels.
    # NOTE: Uses only xmap or ymap (dP[0|1])
    seeds = np.abs(dP[0]) > 1e-3
    if mask is not None:
        seeds = np.logical_or(mask, seeds)

    pixel_loc = np.nonzero(seeds)
    inds = np.array(pixel_loc).astype(np.int32).T

    # Sometimes a random error in empty images.. errr... dunno...
    try:
        p[:, inds[:, 0], inds[:, 1]] = steps2D_interp(
            p=p[:, inds[:, 0], inds[:, 1]],
            dP=dP,
            niter=niter,
            suppress_euler=suppress_euler,
        )
    except Exception:
        pass

    return p, inds



from typing import List

import torch
import torch.nn as nn


class JointLoss(nn.ModuleDict):
    def __init__(
        self,
        losses: List[nn.Module],
        weights: List[float] = None,
    ) -> None:
        """Joint loss function.

        Takes in a list of nn.Module losses and computes the loss for
        each loss in the list and at the end sums the outputs together
        as one joint loss.

        Parameters
        ----------
            losses : List[nn.Module]
                List of initialized nn.Module losses.
            weights : List[float], default=None
                List of weights for each loss.

        Raises
        ------
            ValueError:
                If more than 4 losses are given as input.
                If given weights are not between [0, 1].
        """
        super().__init__()

        if not isinstance(losses, list):
            raise ValueError("`losses` arg must be a list.")

        self.weights = [1.0] * len(losses)
        if weights is not None:
            if not all(0 <= val <= 1.0 for val in weights):
                raise ValueError(f"Weights need to be 0 <= weight <= 1. Got: {weights}")
            self.weights = weights

        for i in range(len(losses)):
            self.add_module(f"loss{i + 1}", losses[i])

    def forward(self, **kwargs) -> torch.Tensor:
        """Compute the joint-loss.

        Returns
        -------
            torch.Tensor:
                The computed joint loss.
        """
        lw = zip(self.values(), self.weights)
        losses = torch.stack([loss(**kwargs) * weight for loss, weight in lw])
        return torch.sum(losses)

    def extra_repr(self) -> str:
        """Add info to print."""
        s = "loss_weights={weights}"
        return s.format(**self.__dict__)



"""
Imported from the cellpose repo. Code style mods.

https://github.com/MouseLand/cellpose/blob/master/cellpose/omnipose/utils.py
https://github.com/MouseLand/cellpose/blob/master/cellpose/omnipose/core.py

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
import skimage.filters as filters

from cellseg_models_pytorch.utils import binarize, percentile_normalize

from .cellpose.integrator import follow_flows
from .cellpose.utils import (
    fill_holes_and_remove_small_masks,
    gen_flows,
    normalize_field,
)

try:
    from sklearn.cluster import DBSCAN
except ImportError:
    raise ImportError(
        """scikit-learn is required for omnipose. `pip install -U scikit-learn`"""
    )


__all__ = ["post_proc_omnipose", "div_rescale", "get_masks_omnipose"]


def div_rescale(dP: np.ndarray, mask: np.ndarray, pad: int = 1) -> np.ndarray:
    """Rescale the divergence of the regressed flows.

    Parameters
    ----------
        dP : np.ndarray
            The regressed eikonal flows. Shape (2, H, W).
        mask : np.ndarray
            Binary mask of the predicted cells. Shape (H, W).
        pad : int, default=1
            Number of padded pixels around the input.

    Returns
    -------
        np.ndarray:
            The rescaled eikonal flow-maps. Shape: (2, H, W).
    """
    dP = dP.copy()
    dP *= mask
    dP = normalize_field(dP)

    # compute the divergence
    y, x = np.nonzero(mask)
    H, W = mask.shape

    Tx = np.zeros((H + 2 * pad) * (W + 2 * pad), np.float64)
    Tx[y * W + x] = np.reshape(dP[1].copy(), H * W)[y * W + x]

    Ty = np.zeros((H + 2 * pad) * (W + 2 * pad), np.float64)
    Ty[y * W + x] = np.reshape(dP[0].copy(), H * W)[y * W + x]

    # Rescaling by the divergence
    div = np.zeros(H * W, np.float64)
    div[y * W + x] = (
        Ty[(y + 2) * W + x]
        + 8 * Ty[(y + 1) * W + x]
        - 8 * Ty[(y - 1) * W + x]
        - Ty[(y - 2) * W + x]
        + Tx[y * W + x + 2]
        + 8 * Tx[y * W + x + 1]
        - 8 * Tx[y * W + x - 1]
        - Tx[y * W + x - 2]
    )

    div.shape = (H, W)
    div = percentile_normalize(div)

    dP *= div

    return dP


def get_masks_omnipose(
    p: np.ndarray,
    mask: np.ndarray,
    inds: np.ndarray = None,
) -> np.ndarray:
    """Omnipose mask recontruction algorithm.

    Parameters
    ----------
        p : np.ndarray
            All the pixel locations for the pixels after running the
            euler integrator. Shape: (2, H, W). Dtype: float32.
        mask : np.ndarray
            The binary mask of the cells. Shape (H, W).
        inds : np.ndarray
            Indices of the non-zero pixels. Shape: (n non zero pxls, 2).

    Returns
    -------
        np.ndarray:
            The instance labelled mask. Shape (H, W).
    """
    eps = 1 + (1 / 3)

    newinds = p[:, inds[:, 0], inds[:, 1]].swapaxes(0, 1)
    mask = np.zeros((p.shape[1], p.shape[2]))

    try:
        db = DBSCAN(eps=eps, min_samples=3, n_jobs=1).fit(newinds)
        labels = db.labels_
        mask[inds[:, 0], inds[:, 1]] = labels + 1
    except Exception:
        pass

    return mask.astype("i4")


def post_proc_omnipose(
    inst_map: np.ndarray,
    flow_map: np.ndarray,
    dist_map: np.ndarray = None,
    return_flows: bool = False,
    min_size: int = 30,
    **kwargs
) -> np.ndarray:
    """Run the omnipose post-processing pipeline.

    More info in the omnipose paper:
    https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled or binary mask. Shape (H, W).
        flow_map : np.ndarray
            Y- and x-flows. Shape: (2, H, W)
        dist_map : np.ndarray, default=None
            Regressed distance transform. Optional. Shape: (H, W).
        return_flows : bool, default=False
            If True, returns the HSV converted flows. They are just not
            needed for anything relevant.
        min_size : int
            The minimum size for the objects that will not be removed.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape (H, W). Dtype: int32.
    """
    #  convert channels to CHW
    if dist_map is not None:
        binary_mask = filters.apply_hysteresis_threshold(dist_map, 0.5, 0.5)
    else:
        binary_mask = binarize(inst_map).astype(bool)

    dP = div_rescale(flow_map, binary_mask)
    pixel_loc, inds = follow_flows(dP, niter=300, mask=binary_mask, suppress_euler=True)

    mask = get_masks_omnipose(p=pixel_loc, mask=binary_mask, inds=inds)

    inst_map = fill_holes_and_remove_small_masks(mask, min_size=min_size).astype("i4")

    if return_flows:
        hsv_flows = gen_flows(dP)
        return inst_map, hsv_flows

    return inst_map



from typing import Dict, Optional, Tuple

import torch
import torch.nn as nn

from ...decoders import Decoder
from ...modules.misc_modules import StyleReshape
from ..base._base_model import BaseMultiTaskSegModel
from ..base._seg_head import SegHead
from ..base._timm_encoder import TimmEncoder
from ._conf import _create_hovernet_args

__all__ = [
    "HoverNet",
    "hovernet_base",
    "hovernet_small",
    "hovernet_plus",
    "hovernet_small_plus",
]


class HoverNet(BaseMultiTaskSegModel):
    def __init__(
        self,
        decoders: Tuple[str, ...],
        heads: Dict[str, Dict[str, int]],
        inst_key: str = "inst",
        depth: int = 4,
        out_channels: Tuple[int, ...] = (512, 256, 64, 64),
        style_channels: int = None,
        enc_name: str = "resnet50",
        enc_pretrain: bool = True,
        enc_freeze: bool = False,
        long_skip: str = "unet",
        merge_policy: str = "sum",
        n_dense: Tuple[int, int] = (8, 4),
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = True,
        attention: str = None,
        preattend: bool = False,
        skip_params: Optional[Dict] = None,
        **kwargs,
    ) -> None:
        """Hover-Net implementation.

        HoVer-Net:
        - https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub

                      (|------ SEMANTIC_DECODER ----- SEMANTIC_HEAD)
                       |
                       |------ TYPE_DECODER ---------- TYPE_HEAD
        ENCODER -------|
                       |------ HOVER_DECODER --------- HOVER_HEAD
                       |
                      (|------ INSTANCE_DECODER ---- INSTANCE_HEAD)

        NOTE: Minor differences from the original implementation.
        - Different encoder, (any encoder from timm-library).
        - Dense blocks have transition conv-blocks like in the original dense-net.

        Parameters
        ----------
            decoders : Tuple[str, ...]
                Names of the decoder branches of this network. E.g. ("hovernet", "sem")
            heads : Dict[str, Dict[str, int]]
                The segmentation heads of the architecture. I.e. Names of the decoder
                branches (has to match `decoders`) mapped to dicts
                of output name - number of output classes. E.g.
                {"hovernet": {"hovernet": 2}, "sem": {"sem": 5}, "type": {"type": 5}}
            inst_key : str, default="inst"
                The key for the model output that will be used in the instance
                segmentation post-processing pipeline as the binary segmentation result.
            depth : int, default=4
                The depth of the encoder. I.e. Number of returned feature maps from
                the encoder. Maximum depth = 5.
            out_channels : Tuple[int, ...], default=(512, 256, 64, 64)
                Out channels for each decoder stage.
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
            enc_name : str, default="resnet50"
                Name of the encoder. See timm docs for more info.
            enc_pretrain : bool, default=True
                Whether to use imagenet pretrained weights in the encoder.
            enc_freeze : bool, default=False
                Freeze encoder weights for training.
            long_skip : str, default="unet"
                long skip method to be used. One of: "unet", "unetpp", "unet3p",
                "unet3p-lite", None
            merge_policy : str, default="sum"
                The long skip merge policy. One of: "sum", "cat"
            n_dense : Tuple[int, int], default=(8, 4)
                Number of dense blocks in the dense decoder stages.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=True
                If True, normalization will be applied before convolution.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
            skip_params : Optional[Dict]
                Extra keyword arguments for the skip-connection module. These depend
                on the skip module. Refer to specific skip modules for more info.

        Raises
        ------
            ValueError: If `decoders` does not contain 'hovernet'.
            ValueError: If `heads` keys don't match `decoders`.
            ValueError: If decoder names don't have a matching head name in `heads`.
        """
        super().__init__()
        self.aux_key = self._check_decoder_args(decoders, ("hovernet",))
        self.inst_key = inst_key
        self._check_head_args(heads, decoders)
        self._check_depth(depth, {"out_channels": out_channels})

        self.enc_freeze = enc_freeze
        use_style = style_channels is not None
        self.heads = heads

        # Create build args
        n_layers = (3, 3) + (1,) * (depth - 2)
        n_blocks = ((1, n_dense[0], 1), (1, n_dense[1], 1)) + ((1,),) * (depth - 2)
        dec_params = {
            d: _create_hovernet_args(
                depth,
                n_dense,
                normalization,
                activation,
                convolution,
                attention,
                preactivate,
                preattend,
                use_style,
                merge_policy,
                skip_params,
            )
            for d in decoders
        }

        # set timm encoder
        self.encoder = TimmEncoder(
            enc_name,
            depth=depth,
            pretrained=enc_pretrain,
        )

        # Style
        self.make_style = None
        if use_style:
            self.make_style = StyleReshape(self.encoder.out_channels[0], style_channels)

        # set decoders and heads
        for decoder_name in decoders:
            decoder = Decoder(
                enc_channels=list(self.encoder.out_channels),
                out_channels=out_channels,
                style_channels=style_channels,
                long_skip=long_skip,
                merge_policy=merge_policy,
                n_layers=n_layers,
                n_blocks=n_blocks,
                stage_params=dec_params[decoder_name],
            )
            self.add_module(f"{decoder_name}_decoder", decoder)

        for decoder_name in heads.keys():
            for output_name, n_classes in heads[decoder_name].items():
                seg_head = SegHead(
                    in_channels=decoder.out_channels,
                    out_channels=n_classes,
                    kernel_size=1,
                )
                self.add_module(f"{output_name}_seg_head", seg_head)

        self.name = f"HoverNet-{enc_name}"

        # init decoder weights
        self.initialize()

        # freeze encoder if specified
        if enc_freeze:
            self.freeze_encoder()

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Forward pass of HoVer-Net."""
        self._check_input_shape(x)

        feats = self.encoder(x)

        style = None
        if self.make_style is not None:
            style = self.make_style(feats[0])

        dec_feats = self.forward_dec_features(feats, style)
        out = self.forward_heads(dec_feats)

        return out


def hovernet_base(type_classes: int, inst_classes: int = 2, **kwargs) -> nn.Module:
    """Create the baseline HoVer-Net (three decoders) from kwargs.

    HoVer-Net:
        - https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub

    Parameters
    ----------
        type_classes : int
            Number of type classes.
        inst_classes : int, default=2
            Number of instance classes.
        **kwargs:
            Arbitrary key word args for the HoverNet class.

    Returns
    -------
        nn.Module: The initialized HoVer-Net model.
    """
    hovernet = HoverNet(
        decoders=("hovernet", "type", "inst"),
        heads={
            "hovernet": {"hovernet": 2},
            "type": {"type": type_classes},
            "inst": {"inst": inst_classes},
        },
        **kwargs,
    )

    return hovernet


def hovernet_plus(
    type_classes: int,
    sem_classes: int,
    inst_classes: int = 2,
    **kwargs,
) -> nn.Module:
    """Create HoVer-Net+ (additional semantic decoders) from kwargs.

    HoVer-Net:
        - https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub

    Parameters
    ----------
        type_classes : int
            Number of type-branch classes.
        sem_classes : int
            Number of semantic-branch classes.
        inst_classes : int, default=2
            Number of instance-branch classes.
        **kwargs:
            Arbitrary key word args for the HoverNet class.

    Returns
    -------
        nn.Module: The initialized HoVer-Net+ model.
    """
    hovernet = HoverNet(
        decoders=("hovernet", "type", "inst", "sem"),
        heads={
            "hovernet": {"hovernet": 2},
            "sem": {"sem": sem_classes},
            "type": {"type": type_classes},
            "inst": {"inst": inst_classes},
        },
        **kwargs,
    )

    return hovernet


def hovernet_small(type_classes: int, **kwargs) -> nn.Module:
    """Create HoVer-Net without inst decoder branch from kwargs.

    HoVer-Net:
        - https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub

    Parameters
    ----------
        type_classes : int
            Number of type-branch classes.
        **kwargs:
            Arbitrary key word args for the HoverNet class.

    Returns
    -------
        nn.Module: The initialized HoVer-Net-small model.
    """
    hovernet = HoverNet(
        decoders=("hovernet", "type"),
        inst_key="type",
        heads={
            "hovernet": {"hovernet": 2},
            "type": {"type": type_classes},
        },
        **kwargs,
    )

    return hovernet


def hovernet_small_plus(type_classes: int, sem_classes: int, **kwargs) -> nn.Module:
    """Create the HoVer-Net+ without inst decoder branch from kwargs.

    HoVer-Net:
        - https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub

    Parameters
    ----------
        type_classes : int
            Number of type-branch classes.
        sem_classes : int
            Number of semantic-branch classes.
        **kwargs:
            Arbitrary key word args for the HoverNet class.

    Returns
    -------
        nn.Module: The initialized HoVer-Net-small+ model.
    """
    hovernet = HoverNet(
        decoders=("hovernet", "type", "sem"),
        inst_key="type",
        heads={
            "hovernet": {"hovernet": 2},
            "type": {"type": type_classes},
            "sem": {"sem": sem_classes},
        },
        **kwargs,
    )

    return hovernet



import pytest
import torch

from cellseg_models_pytorch.utils import (
    dataset_normalize_torch,
    minmax_normalize_torch,
    ndarray_to_tensor,
    normalize_torch,
    percentile_normalize_torch,
)


@pytest.mark.parametrize(
    "norm_method", [minmax_normalize_torch, percentile_normalize_torch, normalize_torch]
)
@pytest.mark.parametrize("batch_dim", [True, False])
def test_channel_norms(img_sample, norm_method, batch_dim):

    if batch_dim:
        img = ndarray_to_tensor(img_sample, "HWC", "BCHW")
    else:
        img = ndarray_to_tensor(img_sample, "HWC", "HWC").permute(2, 0, 1)

    img = norm_method(img)

    assert img.dtype == torch.float32
    assert img.shape == img.shape


@pytest.mark.parametrize("batch_dim", [True, False])
def test_dataset_norm(img_sample, batch_dim):
    if batch_dim:
        img = ndarray_to_tensor(img_sample, "HWC", "BCHW")
    else:
        img = ndarray_to_tensor(img_sample, "HWC", "HWC").permute(2, 0, 1)

    img = dataset_normalize_torch(img, (0.4, 0.5, 0.6), (0.1, 0.2, 0.1), True)

    assert img.dtype == torch.float32
    assert img.shape == img.shape



from typing import Dict, Optional, Tuple

import torch
import torch.nn as nn

from .decoder_stage import DecoderStage

__all__ = ["Decoder"]


class Decoder(nn.ModuleDict):
    def __init__(
        self,
        enc_channels: Tuple[int, ...],
        out_channels: Tuple[int, ...] = (256, 128, 64, 32, 16),
        style_channels: int = None,
        n_layers: Tuple[int, ...] = (1, 1, 1, 1, 1),
        n_blocks: Tuple[Tuple[int, ...], ...] = ((2,), (2,), (2,), (2,), (2,)),
        long_skip: str = "unet",
        stage_params: Optional[Tuple[Dict, ...]] = None,
        **kwargs,
    ) -> None:
        """Build a generic U-net-like decoder.

        Parameters
        ----------
            enc_channels : Tuple[int, ...]
                Number of channels at each encoder layer.
            out_channels : Tuple[int, ...], default=(256, 128, 64, 32, 16)
                Number of channels at each decoder layer output.
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
            n_layers : Tuple[int, ...], default=(1, 1, 1, 1, 1)
                The number of conv layers inside each of the decoder stages.
            n_blocks : Tuple[Tuple[int, ...], ...] = ((2, ), (2, ), (2, ). (2, ), (2, ))
                The number of blocks inside each conv-layer in each decoder stage.
            long_skip : str, default="unet"
                long skip method to be used. One of: "unet", "unetpp", "unet3p",
                "unet3p-lite", None
            stage_params : Optional[Tuple[Dict, ...]], default=None
                The keyword args for each of the distinct decoder stages. Incudes the
                parameters for the long skip connections and convolutional layers of the
                decoder itself. See the `DecoderStage` documentation for more info.
        """
        super().__init__()
        self.long_skip = long_skip

        if not len(out_channels) == len(enc_channels):
            raise ValueError(
                f"""The number of encoder channels need to match the number of
                decoder channels. Got {len(out_channels)} decoder channels
                and {len(enc_channels)} encoder channels"""
            )

        out_channels = [enc_channels[0]] + list(out_channels)
        skip_channels = enc_channels[1:]

        # scaling factor assumed to be 2 for the spatial dims and the input
        # has to be divisible by 32. 256 used here just for convenience.
        depth = len(out_channels)
        out_dims = [256 // 2**i for i in range(depth)][::-1]

        # Build decoder
        for i in range(depth - 1):
            decoder_block = DecoderStage(
                stage_ix=i,
                dec_channels=tuple(out_channels),
                dec_dims=tuple(out_dims),
                skip_channels=skip_channels,
                style_channels=style_channels,
                long_skip=long_skip,
                n_layers=n_layers[i],
                n_blocks=n_blocks[i],
                **stage_params[i] if stage_params is not None else {"k": None},
            )
            self.add_module(f"decoder_stage{i + 1}", decoder_block)

        self.out_channels = decoder_block.out_channels

    def forward(
        self, *features: Tuple[torch.Tensor], style: torch.Tensor = None
    ) -> torch.Tensor:
        """Forward pass of the decoder."""
        head = features[0]
        skips = features[1:]
        extra_skips = [head] if self.long_skip == "unet3p" else []

        x = head
        for _, decoder_stage in enumerate(self.values()):
            x, extra = decoder_stage(
                x, skips=skips, extra_skips=extra_skips, style=style
            )

            if self.long_skip == "unetpp":
                extra_skips = extra
            elif self.long_skip == "unet3p":
                extra_skips.append(x)

        return x



from pathlib import Path
from typing import Callable, List, Optional, Tuple

import numpy as np
from tqdm import tqdm

from ...utils import FileHandler
from ._base_writer import BaseWriter

__all__ = ["SlidingWindowFolderWriter"]


class SlidingWindowFolderWriter(BaseWriter):
    def __init__(
        self,
        in_dir_im: str,
        in_dir_mask: str,
        save_dir_im: str,
        save_dir_mask: str,
        patch_size: Tuple[int, int],
        stride: int,
        transforms: Optional[List[str]] = None,
    ) -> None:
        """Write overlapping patches to a folder from image and .mat files.

        Image patches will be written as .png and mask patches as .mat files.

        Parameters
        ----------
            in_dir_im : str
                Path to the folder of images
            in_dir_mask : str
                Path to the folder of masks.
            save_dir_im : str
                Path to the folder where the new img patches will be saved.
            save_dir_mask : str
                Path to the folder where the new mask patches will be saved.
            patch_size : Tuple[int, int]
                Height and width of the extracted patches.
            stride : int
                Stride for the sliding window.
            transforms : List[str]:
                A list of transforms to apply to the images and masks. Allowed ones:
                "blur", "non_spatial", "non_rigid", "rigid", "hue_sat", "random_crop",
                "center_crop", "resize"

        Raises
        ------
            ValueError if issues with given paths or the files in those folders.

        Example
        -------
            >>> writer = SlidingWindowFolderWriter(
                    "/path/to/my/imgs/,
                    "/path/to/my/masks/,
                    "/save/img_patches/here/,
                    "/save/mask_patches/here/,
                    patch_size=(320, 320),
                    stride=160,
                    transforms=["rigid"]
                )
            >>> writer.write()

        """
        super().__init__(
            in_dir_im=in_dir_im,
            in_dir_mask=in_dir_mask,
            stride=stride,
            patch_size=patch_size,
            transforms=transforms,
        )
        self.save_dir_im = Path(save_dir_im)
        self.save_dir_mask = Path(save_dir_mask)

    def write(self, pre_proc: Callable = None, msg: str = None) -> None:
        """Write patches to a folder.

        Parameters
        ----------
            pre_proc : Callable, optional
                An optional pre-processing function for the masks.
        """
        with tqdm(
            zip(self.fnames_im, self.fnames_mask), total=len(self.fnames_im)
        ) as pbar:
            total_tiles = 0
            for fni, fnm in pbar:
                msg = msg if msg is not None else ""
                pbar.set_description(f"Extracting {msg} patches to folders..")
                tiles = self._get_tiles(fni, fnm, pre_proc)
                n_tiles = tiles["image"].shape[0]

                arg_list = []
                for i in range(n_tiles):
                    fn = fni.name[:-4]
                    dd = dict(
                        path_im=self.save_dir_im / f"{fn}_patch{i+1}.png",
                        path_mask=self.save_dir_mask / f"{fn}_patch{i+1}.mat",
                    )
                    for k, m in tiles.items():
                        dd[k] = m[i : i + 1].squeeze()
                    arg_list.append(dd)
                self._write_parallel(self._save2folder, arg_list)
                total_tiles += n_tiles
                pbar.set_postfix_str(f"# of extracted tiles {total_tiles}")

    def save2folder(
        self,
        path_im: str,
        path_mask: str,
        image: np.ndarray,
        inst_map: np.ndarray,
        type_map: np.ndarray = None,
        sem_map: np.ndarray = None,
    ) -> None:
        """Write image and corresponding masks to folder."""
        FileHandler.write_img(path_im, image)
        FileHandler.write_mask(path_mask, inst_map, type_map, sem_map)

    def _save2folder(self, kwargs) -> None:
        """Unpack the kwargs for `save2folder`."""
        return self.save2folder(**kwargs)
